/*
 *  This file is part of the X10 project (http://x10-lang.org).
 *
 *  This file is licensed to You under the Eclipse Public License (EPL);
 *  You may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *      http://www.opensource.org/licenses/eclipse-1.0.php
 *
 *  (C) Copyright IBM Corporation 2006-2016.
 */

package apgas.impl;

import static apgas.Constructs.here;

import apgas.DeadPlaceException;
import apgas.Job;
import apgas.SerializableJob;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.ForkJoinTask;
import java.util.concurrent.RecursiveAction;

/**
 * The {@link Task} class represents an APGAS task.
 *
 * <p>This class implements task serialization and handles errors in the serialization process.
 */
public final class Task extends RecursiveAction implements SerializableRunnable {

  private static final long serialVersionUID = 5288338719050788305L;

  private static final SerializableJob NULL = () -> {};
  /** The finish object for this {@link Task} instance. */
  Finish finish;

  /**
   * Extra Finish instances in which this instance is registered.
   * <p>
   * This array will be of size 0 under the "normal" finish/async model. Only when
   * the
   * {@link apgas.ExtendedConstructs#asyncAtWithCoFinish(apgas.Place, SerializableJob, Finish...)}
   * is called will this array be populated.
   */
  Finish[] coFinish;

  /** System wide unique id of this task */
  /** The place of the parent task. */
  private int parent;

  /** The function to run. */
  private Job f;

  /**
   * Constructs a new {@link Task}.
   *
   * @param f the function to run
   * @param finish the finish object for this task
   * @param parent the place of the parent task
   */
  Task(Finish finish, Job f, int parent, Finish... extraFinishes) {
    this.finish = finish;
    this.f = f;
    this.parent = parent;
    coFinish = extraFinishes;
  }

  public Finish getFinish() {
    return finish;
  }

  /** Submits the task for asynchronous execution. */
  @Override
  public void run() {
    try {
      async(null);
    } catch (final DeadPlaceException e) {
      // source place has died while task was in transit, discard
    }
  }

  /** Runs the task and notify the task's finish upon termination. */
  @Override
  protected void compute() {
    final Worker worker = (Worker) Thread.currentThread();
    worker.task = this;
    try {
      f.run();
    } catch (final Throwable t) {
      System.out.println("[APGAS] " + here() + " caught Exception");
      t.printStackTrace(System.out);
      finish.addSuppressed(t);
    }
    finish.tell();
  }

  /**
   * Submits the task for asynchronous execution.
   *
   * @param worker the worker doing the submission or null if not a worker thread
   */
  void async(Worker worker) {
    finish.submit(parent);

    if (worker == null) {
      GlobalRuntimeImpl.getRuntime().execute(this);
    } else {
      fork();
    }
  }

  /**
   * Runs the tasks, notify the task's finish upon termination, and wait for the task's finish to
   * terminate.
   *
   * @param worker the worker thread running the task or null if not a worker thread
   */
  void finish(Worker worker) {
    if (worker == null) {
      async(null);
      try {
        ForkJoinPool.managedBlock(finish);
      } catch (final InterruptedException e) {
      }
    } else {
      final Task savedTask = worker.task;
      compute();
      Task t;
      while (!finish.isReleasable()
          && (t = (Task) ForkJoinTask.peekNextLocalTask()) != null
          && finish == t.finish
          && t.tryUnfork()) {
        t.compute();
      }
      try {
        ForkJoinPool.managedBlock(finish);
      } catch (final InterruptedException e) {
      }
      worker.task = savedTask;
    }
  }

  /**
   * Submits the task for asynchronous execution at place p.
   *
   * <p>If serialization fails, the task is dropped and the task's finish is notified. The exception
   * is logged to System.err and masked unless APGAS_SERIALIZATION_EXCEPTION is set to "true".
   *
   * @param p the place ID
   */
  void asyncAt(int p) {
    try {
      GlobalRuntimeImpl.getRuntime().transport.send(p, this);
    } catch (final Throwable e) {
      finish.unspawn(p);
      throw e;
    }
  }

  /**
   * Serializes the task.
   *
   * @param out the object output stream
   * @throws IOException if I/O errors occur
   */
  private void writeObject(ObjectOutputStream out) throws IOException {
    out.writeObject(finish);
    out.writeInt(parent);
    out.writeObject(f);
  }

  /**
   * Deserializes the task.
   *
   * <p>If deserialization fails, a dummy task is returned. The exception is logged to System.err
   * and masked unless APGAS_SERIALIZATION_EXCEPTION is set to "true".
   *
   * @param in the object input stream
   * @throws IOException if I/O errors occur
   * @throws ClassNotFoundException if the class of the serialized object cannot be found
   */
  private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
    finish = (Finish) in.readObject();
    parent = in.readInt();
    try {
      f = (SerializableJob) in.readObject();
    } catch (final Throwable e) {
      finish.addSuppressed(e);
      f = NULL;
    }
  }
}
