package nfjglbsw.examples.fib;

import java.io.Serializable;
import java.util.Iterator;

import nfjglbsw.examples.util.LongSum;
import nfjglbsw.multiworker.Bag;
import nfjglbsw.multiworker.GLBWorker;
import nfjglbsw.multiworker.GLBcomputer;
import nfjglbsw.multiworker.NFJTask;

public class Fib implements Bag<Fib, LongSum>, Serializable {

  public long cutOff;
  public FibTaskQueue queue;

  public Fib(long cutOff) {
    this(cutOff, 32);
  }

  public Fib(long cutOff, int initialQueueSize) {
    this.queue = new FibTaskQueue(initialQueueSize);
    this.cutOff = cutOff;
  }

  public void init(int n, GLBcomputer<LongSum, Fib> computer) {
    this.queue.addFirst(new FibTask(n, null, 0));
  }

  @Override
  public void initStaticTasks(int workerId) {
    // unused because dynamic
  }

  @Override
  public boolean isEmpty() {
    return this.queue.size() == 0;
  }

  @Override
  public boolean isSplittable() {
    for (FibTask task : queue) {
      if (!task.reachedJoin()) {
        return this.queue.size() > 1;
      }
    }
    return false;
  }

  @Override
  public void merge(Fib other) {
    this.queue.pushArrayFirst(other.queue.toArray());
  }

  @Override
  public void merge(NFJTask task) {
    this.queue.addFirst((FibTask) task);
  }

  @Override
  public NFJTask[] asTasks() {
    NFJTask[] tasks = new NFJTask[this.queue.size()];
    int i = 0;
    for (NFJTask task : this.queue) {
      tasks[i++] = task;
    }
    return tasks;
  }

  @Override
  public NFJTask process(int workAmount, LongSum sharedObject, GLBWorker<LongSum, Fib> computer) {
//    System.out.println(computer.workerID + " entering process q.size = " + queue.size());
    int i = 0;
    if (this.queue.size() == 0) return null;
    FibTask task = this.queue.removeFirst();
    while (i < workAmount) {
/*      System.out.println(
          computer.workerID
              + " processes frame with cont="
              + task.continuation
              + " with n="
              + task.n
              + " q.size= "
              + queue.size());*/
      ++i;
      switch (task.continuation) {
        case 0:
          if (task.n <= cutOff) {
            computer.parentReturn(task, fibonacci(task.n));
            task = null;
          } else {
            task.continuation = 1;
            this.queue.addFirst(task);
            final FibTask parentTask = task;
            task = new FibTask(task.n - 2, parentTask, computer.generateNewTaskID());
            parentTask.childID1 = task.getID();
            continue;
          }
          break;
        case 1:
          task.continuation = 2;
          this.queue.addFirst(task);
          final FibTask parentTask = task;
          task = new FibTask(task.n - 1, parentTask, computer.generateNewTaskID());
          parentTask.childID2 = task.getID();
          continue;
        case 2: // join
          task.continuation = 3;
/*          System.out.println(
              computer.workerID
                  + " joining task cont="
                  + task.continuation
                  + " n="
                  + task.n
                  + " q.size= "
                  + queue.size());*/
          computer.join(task, task.childID1, task.childID2);
          task = null;
          break;
        case 3: // return
          Long result = task.getChildResult(task.childID1) + task.getChildResult(task.childID2);
          if (task.getParent() == null) {
            System.out.println(
                computer.workerID
                    + " RESULT IS "
                    + result
                    + " (FROM TASK ID = "
                    + task.getID()
                    + " n = "
                    + task.n
                    + " cont="
                    + task.continuation
                    + ")");
            if (result != fastFib(task.n)) {
              System.out.println("RESULT NOT CORRECT (fastFib(" + task.n + ") = " + fastFib(task.n) + "!");
            }
          }

          computer.parentReturn(task, result);
          task = null;
          break;
      }
      if (this.queue.size() > 0 && i < workAmount) {
        task = this.queue.removeFirst();
      } else {
        break;
      }
    }
    return task;
  }

  public Fib split(boolean takeAll, int numTasks) {
    Fib other = new Fib(cutOff, 1);

    if (!isSplittable() && takeAll) {
      other.queue.pushArrayFirst(this.queue.getFromLast(this.queue.size()));
      return other;
    }

    int joined = 0;
    for (Iterator<FibTask> iterator = queue.descendingIterator(); iterator.hasNext(); ) {
      FibTask task = iterator.next();
      if (!task.reachedJoin()) {
        FibTask[] fromLast = this.queue.getFromLast(joined);
        other.queue.addLast(this.queue.removeLast());
        this.queue.pushArrayLast(fromLast);
        break;
      } else {
        joined++;
      }
    }
    return other;
  }

  @Override
  public long getCurrentTaskCount() {
    return this.queue.size();
  }

  private static long fibonacci(int a) {
    if (a == 1 || a == 2) return 1L;
    else return fibonacci(a - 1) + fibonacci(a - 2);
  }

  private static long fastFib(int a) {
    long n1 = 1;
    long n2 = 1;
    long n3 = 0;
    for (int i = 3; i <= a; ++i) {
      n3 = n1 + n2;
      n1 = n2;
      n2 = n3;
    }
    return n3;
  }

  @Override
  public Fib deepCopy() {
    Fib other = new Fib(cutOff);
    for (FibTask task : this.queue) {
      other.queue.addLast(new FibTask(task));
    }
    return other;
  }
}
