package nfjglbsw.multiworker;

import static apgas.Constructs.*;

import apgas.util.ConsolePrinter;
import java.io.Serializable;
import java.util.*;
import java.util.concurrent.*;

import nfjglbsw.multiworker.lifeline.LifelineStrategy;

/**
 * Class {@link GLBWorker} implements a lifeline-based work-stealing scheme between hosts with an
 * internal management that allows multiple workers to perform computation concurrently on the same
 * Java process.
 *
 * <p>This {@link GLBWorker} handles distributed computation that fits the {@link Bag} interface
 * using multiple concurrent workers on the same hosts. It follows the design proposed by Yamashita
 * and Kamada in their article <a href=
 * "https://www.jstage.jst.go.jp/article/ipsjjip/24/2/24_416/_article">Introducing a Multithread and
 * Multistage Mechanism for the Global Load Balancing Library of X10</a>. The current design does
 * not implement the multi-stage mechanisms described in their article.
 *
 * <p>The requirements on the kind of computation and the features available to the programmer are
 * further detailed in the documentation of classes {@link Bag} and {@link Fold}. The customization
 * possibilities of the load balance algorithm are presented in class {@link
 * GLBMultiWorkerConfiguration}.
 *
 * <p>The work-stealing scheme is similar in spirit to the original lifeline-based Global Load
 * Balancer implemented in the X10 programming language. Inactive hosts passively wait for some work
 * to reach them through their lifelines to resume computation. The major difference with the
 * original scheme comes in the fact that this implementation accommodates for several concurrent
 * workers in a single Java process. Some load balance is performed internally to keep the workers
 * occupied as much as possible while still allowing remote hosts to steal work. The design relies
 * on two {@link Bag} instances that are kept aside to perform load balance. One is primarily in
 * charge of load balance operations between the workers running on the local host (member {@link
 * #intraPlaceQueue}) while the other in dedicated to steals from remote hosts (member {@link
 * #interPlaceQueue}).
 *
 * @author Patrick Finnerty
 */

/** <R> type of the result produced by the computation <B> type of the computation bag */
public class GLBWorker<R extends Fold<R> & Serializable, B extends Bag<B, R> & Serializable> {

  /** Holds the lifelineStrategy */
  private final LifelineStrategy lifelineStrategy;
  /** Printing Helper */
  ConsolePrinter console = ConsolePrinter.getInstance();

  /**
   * Lifelines this place can establish.
   */
  int[] LIFELINE;
  /**
   * Places that can establish a lifeline on this place.
   */
  int[] REVERSE_LIFELINE;
  /**
   * {@link Logger} instance used to gather the {@link PlaceLogger}s of each place and hold runtime
   * information common to the whole computation.
   *
   * @see #getLog()
   */
  Logger computationLog;
  /**
   * Collection used to keep track of the lifelines this place has established on other places.
   *
   * <p>The key is the integer identifying the remote place on which this place may establish
   * lifelines, the value is {@code true} when the lifeline is established, {@code false} otherwise.
   * There is always a mapping in this member for every potential lifeline established by this
   * place.
   */
  ConcurrentHashMap<Integer, Boolean> lifelineEstablished;
  /**
   * Collection used to record the lifeline thieves that have requested some work from this place
   * but could not be given some work straight away as part of method steal.
   */
  ConcurrentLinkedQueue<Integer> lifelineThieves;
  /**
   * Logger instance used to log the runtime of the {@link GLBWorker} instance located at this place
   */
  PlaceLogger logger;
  /**
   * Indicates if the log aggregation has already being performed in method {@link #getLog()}.
   * Avoids a second aggregation of the logs if multiple calls to this method are made following a
   * computation.
   */
  boolean logsGiven;
  /** Random instance used to decide the victims of random steals. */
  Random random;
  /**
   * Instance in which the result of the computation performed at this place is going to be stored.
   * It is initialized with the given neutral element before the computation starts in method {@link
   * #reset(SerializableSupplier, SerializableSupplier, SerializableSupplier)}.
   */
  R result;
  /**
   * State of this worker.
   *
   * <ul>
   *   <li>0 running
   *   <li>-1 stealing
   *   <li>-2 inactive
   * </ul>
   */
  volatile int state;

  volatile WorkerBag workerBag;

  /** Global ID of this worker. */
  public final int workerID;

  private SerializableSupplier<R> resultInitializer;
  private SerializableSupplier<B> queueInitializer;
  private SerializableSupplier<B> workerInitializer;

  volatile CountDownLatch waitLatch;

  ConcurrentHashMap<Long, NFJTask>
      givenFrames; // frames that were given away to a thief (TaskID -> Task)
  ConcurrentHashMap<Long, Object>
      storedResults; // results not yet inserted into the parent frame (TaskID -> TaskResult)
  //  ConcurrentHashMap<Long, NFJTask>
  //      localFrames; // frames that were spawned on this place (not necessarily still on this
  // place)
  // (TaskID -> Task)
  ConcurrentHashMap<Long, ArrayList<Long>>
      pendingFrames; // frames that wait for task results (TaskID -> List(ChildTaskIDs))
  ConcurrentHashMap<Long, ArrayList<Long>>
      pendingFramesLocal; // frames that wait for local task results (TaskID -> List(ChildTaskIDs))

  ConcurrentLinkedQueue<Integer> randomStealRequests;

  volatile long lastTaskID;

  public long startTime;

  public long processedTime;

  volatile long openTasks;
  volatile boolean latestWasSplittable;

  int localoptMode;

  public int[] randomWorkerArray;
  public int lastRandomWorkerIdx;

  public void setLastTaskID(long lastTaskID) {
    this.lastTaskID = lastTaskID;
  }

  public long generateNewTaskID() {
    return ++lastTaskID;
  }

  public boolean isLocalTask(long taskID) {
    return (workerID == (taskID >> 32L));
  }

  public int getNextRandomVictim() {
    lastRandomWorkerIdx = (lastRandomWorkerIdx + 1) % randomWorkerArray.length;
    return randomWorkerArray[lastRandomWorkerIdx];
  }

  GLBcomputer<R, B> computer;

  /** Constructor (package visibility) */
  GLBWorker(int workerID, GLBcomputer computer, int totalWorkers) {
    this.startTime = System.currentTimeMillis();
    this.workerID = workerID;
    this.lastTaskID = 1 + (((long) workerID) << 32L);
    this.computer = computer;

    LifelineStrategy ls = null;
    try {
      System.out.println(GLBMultiWorkerConfiguration.GLB_MULTIWORKER_LIFELINESTRATEGY.get());
      ls =
          (LifelineStrategy)
              Class.forName(GLBMultiWorkerConfiguration.GLB_MULTIWORKER_LIFELINESTRATEGY.get())
                  .newInstance();
    } catch (InstantiationException | IllegalAccessException | ClassNotFoundException e) {
      e.printStackTrace();
    }

    this.localoptMode = GLBMultiWorkerConfiguration.GLB_MULTIWORKER_LOCALOPT.get();

    this.lifelineStrategy = ls;

    LIFELINE = lifelineStrategy.lifeline(workerID, totalWorkers);
    REVERSE_LIFELINE = lifelineStrategy.reverseLifeline(workerID, totalWorkers);

    random = new Random(workerID);

    lifelineEstablished = new ConcurrentHashMap<>(LIFELINE.length);
    lifelineThieves = new ConcurrentLinkedQueue<>();
    logger = new PlaceLogger(workerID);

    storedResults = new ConcurrentHashMap<>();
    givenFrames = new ConcurrentHashMap<>();
    pendingFrames = new ConcurrentHashMap<>();
    pendingFramesLocal = new ConcurrentHashMap<>();

    randomStealRequests = new ConcurrentLinkedQueue<>();

    // generate array containing all workers exactly once
    randomWorkerArray = new int[totalWorkers - 1];
    for (int i = 0; i < randomWorkerArray.length; ++i) {
      randomWorkerArray[i] = -1;
    }
    if (totalWorkers > 1) {
      for (int i = 0; i < totalWorkers; ++i) {
        if (i == workerID) continue;
        int idx = 0;
        do {
          idx = random.nextInt(randomWorkerArray.length);
        } while (randomWorkerArray[idx] != -1);
          randomWorkerArray[idx] = i;
      }
      lastRandomWorkerIdx = -1;
    }
  }

  /**
   * Computes the given bag and returns the aggregated result of this computation.
   *
   * @param bag the computation to be performed
   * @param initResultSupplier function that provides new empty result instances
   * @param emptyBagSupplier function that provides new empty computation bag instances
   * @return aggregated result of the computation
   */
  public R computeDynamic(
      B bag, SerializableSupplier<R> initResultSupplier, SerializableSupplier<B> emptyBagSupplier) {
    return computeDynamic(bag, initResultSupplier, emptyBagSupplier, emptyBagSupplier);
  }

  /**
   * Computes the given bag and returns the result of the distributed computation.
   *
   * @param work initial work to be processed
   * @param resultInitializer initializer for the result instance
   * @param queueInitializer initializer for the queue used for load balancing purposes
   * @param workerInitializer initializer for the workers bag
   * @return instance of type R containing the result of the distributed computation
   */
  public R computeDynamic(
      B work,
      SerializableSupplier<R> resultInitializer,
      SerializableSupplier<B> queueInitializer,
      SerializableSupplier<B> workerInitializer) {
    // We reset every place
    final long initStart = System.nanoTime();
    this.resultInitializer = resultInitializer;
    this.queueInitializer = queueInitializer;
    this.workerInitializer = workerInitializer;

    // We launch the computation
    setState(0);
    final long start = System.nanoTime();
    try {
      run(work);
    } catch (Throwable t) {
      console.println("Exception caught");
      t.printStackTrace(System.out);
    }
    console.println("after finish run");

    final long computationFinish = System.nanoTime();
    System.out.println("Process time: " + ((computationFinish - start) / 1E9) + " s");

    // Preparation for method getLog if it is called
    this.computationLog.setTimings(initStart, start, computationFinish);
    return result;
  }

  /**
   * Method called on this worker when a victim of steal is answering and providing some loot.
   *
   * @param victim the id from worker sending the loot
   * @param random true if the steal is a random steal, otherwise false
   * @param loot the work that was stolen by this worker
   */
  void deal(int victim, boolean random, B loot) {
    if (random) {
      logger.stealsSuccess.incrementAndGet();
    } else {
      logger.lifelineStealsSuccess.incrementAndGet();
      lifelineEstablished.put(victim, false);
    }

    if (computer.isLocalWorker(victim)) {
      loot = loot.deepCopy();
    }

    for (NFJTask task : loot.asTasks()) {
      task.victims.add(victim);
      task.setStolen(false);
      if (givenFrames.containsKey(task.getID())) {
        //      task.parentTask = null;
        NFJTask givenTask = givenFrames.get(task.getID());
        givenFrames.put(task.getID(), task);
        task.parentTask = givenTask.parentTask;
      }
      if (task.getID() == 0) console.println(workerID + " root task moved (deal) from " + victim + " -> " + workerID);
    }

    console.println(
        workerID + " MERGE LOOT INTO WORKER "
            + " lootsize="
            + loot.getCurrentTaskCount()
            + " q.size = "
            + workerBag.bag.getCurrentTaskCount()
            + " timestamp = " + (System.currentTimeMillis() - startTime));

    synchronized (workerBag) {
      if (state == -2) {
        setState(0);
        console.println(workerID + " start worker random=" + random);
        B finalLoot = loot;
        async(
            () -> {
              run(finalLoot);
            });
      } else {
        workerBag.bag.merge(loot);
      }
    }

    if (waitLatch != null && random == true) {
      waitLatch.countDown();
    }
    console.println(workerID + " end of deal with victim = " + victim + " random=" + random);
  }

  public Logger getLog() {
    if (!logsGiven) {
      try {
        finish(
            () -> {
              /*              for (final Place p : places()) {
                asyncAt(
                    p,
                    () -> {
                      final PlaceLogger l = logger;
                      asyncAt(
                          place(0),
                          () -> {
                            computationLog.addPlaceLogger(l);
                          });
                    });
              }*/
            });
      } catch (Throwable t) {
        console.println("Exception caught");
        t.printStackTrace(System.out);
      }
      logsGiven = true;
    }
    return computationLog;
  }

  /**
   * Resets the local GLBWorker instance to a ready to compute state in preparation for the next
   * computation. An instance of the result which the computation is going to produce is kept aside
   * for later use.
   *
   * <p>
   *
   * @param resultInitSupplier supplier of empty result instance
   * @param queueInitializer supplier of empty queues for load balancing purposes
   * @param workerInitializer supplier of empty bag for workers
   */
  void reset(
      SerializableSupplier<R> resultInitSupplier,
      SerializableSupplier<B> queueInitializer,
      SerializableSupplier<B> workerInitializer) {

    // Resetting the logger
    logger = new PlaceLogger(workerID);
    logsGiven = false;
    computationLog = new Logger();

    // Resetting the field used to keep the result
    result = resultInitSupplier.get();

    this.resultInitializer = resultInitSupplier;
    this.queueInitializer = queueInitializer;
    this.workerInitializer = workerInitializer;

    // Resetting flags
    setState(-2);

    workerBag = new WorkerBag(workerID, workerInitializer);

    // We reset the established lifelines trackers
    final boolean lifelinesOn = (workerID != 0);
    for (final int i : LIFELINE) {
      lifelineEstablished.put(i, lifelinesOn);
    }
    console.println("lifelineEstablished=" + lifelineEstablished);

    // We establish lifelines on this place for initial work-stealing conditions
    for (final int i : REVERSE_LIFELINE) {
      // i==0 is normal case
      if (i == 0) {
        continue;
      }
      lifelineThieves.add(i);
    }
    console.println("lifelineThieves=" + lifelineThieves);
  }

  /**
   * Main procedure of a place.
   *
   * <p>Spawns the first worker thread (which will in turn recursively spawn other worker threads).
   * When all workers run out of work, attempts to steal work from remote hosts.
   *
   * @param b the initial bag to compute.
   */
  void run(B b) {
    console.println("new worker starts");

    // Prepare the first worker to process the work given as parameter
    if (b != null) { // dynamic computation
      console.println(workerID + " trying to enter synchronized in run timestamp = " + (System.currentTimeMillis() - startTime));
      synchronized (workerBag) {
        console.println(
            "Old bag = "
                + workerBag.bag.getCurrentTaskCount()
                + " b.size = "
                + b.getCurrentTaskCount() + " timestamp = " + (System.currentTimeMillis() - startTime));
        workerBag.bag.merge(b);
      }
    }
    //    console.println("before double do loop");

    setState(0);
    try {
      finish(
          () -> {
            // Spawn a first worker (which will spawn the others)
            async(
                () -> {
                  workerProcess(workerBag); // Working
                });
          });
    } catch (Throwable t) {
      console.println("Exception caught");
      t.printStackTrace(System.out);
    }
    // All the workers have stopped, this place does not have any work and
    // will now attempt to steal some from other places

    synchronized (workerBag) {
      console.println(
          workerID + " q.size = " + workerBag.bag.getCurrentTaskCount() + " state = " + state);
      console.println(
          workerID
              + " pending frames size "
              + this.pendingFrames.size()
              + " pending local frames size = "
              + this.pendingFramesLocal.size());
      console.println(
          workerID
              + " givenFrames="
              + this.givenFrames.size()
              + " storedResults.size()= "
              + this.storedResults.size());
      if (pendingFrames.size() > 0) {
        for (Long taskID : pendingFrames.keySet()) {
          ArrayList<Long> childIDs = pendingFrames.get(taskID);
          String output = taskID + ": ";
          for (long child : childIDs) {
            output += child + ", ";
          }
          console.println(workerID + " " + output);
        }
      }
      if (storedResults.size() > 0) {
        console.println("Stored Results: ");
        for (Long taskID : storedResults.keySet()) {
          Object result = storedResults.get(taskID);
          String output = taskID + ": " + result;
          console.println(workerID + " " + output);
        }
      }
    }

    console.println("going to sleep");
  }

  // sends steal requests to a number of random victims
  // waits for the answers (either reject or loot)
  boolean doRandomSteals() {
    console.println(workerID + " beginning random steals");
    if (computer.getNumWorkers() == 1) return false;
    final int numTries = GLBMultiWorkerConfiguration.GLB_MULTIWORKER_W.get();
    final int thief = workerID;
    for (int i = 0; i < numTries; ++i) {
      int victim = -1;
      if (localoptMode > 0 && i == 0) {
        victim = computer.getBestLocalWorker();
      }
      if (victim == -1) {
        victim = getNextRandomVictim();
      }

      if (waitLatch != null) {
        console.println(workerID + " wait latch not NULL!");
        System.exit(22);
        waitLatch.countDown();
      }
      waitLatch = new CountDownLatch(1);
      try {
        final int _victim = victim;
        computer.uncountedAsyncAt(
            victim,
            (GLBWorker worker) -> {
              worker.computer.processIncomingSteal(_victim, thief, false);
            });
      } catch (Throwable t) {
        t.printStackTrace(System.out);
      }

      long waitedFrom = System.currentTimeMillis();
      console.println(workerID + " waiting for victim = " + victim);
      try {
        waitLatch.await();
      } catch (InterruptedException e) {
        e.printStackTrace();
      }
      long waitedTo = System.currentTimeMillis();
      console.println(workerID + " finished waiting for victim = " + victim + " duration = " + (waitedTo - waitedFrom) + " ms");
      waitLatch = null;
      console.println(workerID + " waitLatch set to null");

      // Checks if we have some work now
      boolean workAvailable;
      synchronized (this.workerBag) {
       workAvailable = !workerBag.bag.isEmpty();
      }
      if (workAvailable) {
        setState(0);
        console.println(workerID + " ending random steals successful!");
        return true;
      }
    }
    console.println(workerID + " ending random steals UNsuccessful!");
    return false;
  }

  // sends lifeline steal requests to this worker's lifeline buddies
  // does not wait
  void doLifelineSteals() {
    if (computer.getNumWorkers() == 1) return;
    final int thief = workerID;
    for (final int victim : LIFELINE) {
      computer.asyncAt(
          victim,
          (GLBWorker worker) -> {
            worker.computer.processIncomingSteal(victim, thief, true);
          });
    }
  }

  long lastStateBegin;
  void setState(int state) {
    console.println(workerID + " was on state " + this.state + " (new state=" + state + ") for " + (System.currentTimeMillis() - lastStateBegin) + " ms");
    lastStateBegin = System.currentTimeMillis();
//    System.out.println(workerID + " setting state to " + state + " stackTrace = " + Arrays.toString(Thread.currentThread().getStackTrace()));
    this.state = state;
  }

  void processIncomingSteal(int thief, boolean lifeline) {
    console.println(
        workerID + " processing incoming steal from " + thief + " lifeline = " + lifeline + " current state =" + state + " and waitLatch = " + waitLatch + " timestamp = " + (System.currentTimeMillis() - startTime) + " ms");
    final int victim = workerID;
    if (lifeline) {
      lifelineThieves.add(thief);
    } else {
      console.println(
          workerID + " trying to enter synchronized thief = " + thief + " lifeline = " + lifeline);
      try {
        if (state != 0) { // reject
          console.println(workerID + " rejecting thief = " + thief + " lifeline = " + lifeline);
          computer.uncountedAsyncAt(
                  thief,
                  (GLBWorker worker) -> {
                    worker.console.println(worker.workerID + " received reject message for steal, countdown");
                    worker.waitLatch.countDown();
                  });
        } else {
          randomStealRequests.add(thief);
          console.println(workerID + " Recording random steal request from thief = " + thief);
        }
      } catch (Exception e) {
        e.printStackTrace(System.out);
      }
    }
  }

  public void join(NFJTask task, long... childIDs) {
    ArrayList<Long> childList = new ArrayList();
    for (int i = 0; i < childIDs.length; ++i) {
      childList.add(childIDs[i]);
    }
    join(task, childList);
  }

  // this method returns immediately
  // the task will get scheduled on the first victim of this frame
  public void join(NFJTask task, ArrayList<Long> childIDs) {
    //System.out.println(workerID + " join task.id = " + task.id);
    // fill in locally calculated child results
    ArrayList<Long> pendingChildIDs = new ArrayList<>();
    ArrayList<Long> toRemove = new ArrayList<>();
    synchronized (this.pendingFrames) {
      for (long childID : childIDs) {
        if (this.storedResults.containsKey(childID)) {
          task.receiveChildResult(childID, this.storedResults.remove(childID));
          toRemove.add(childID);
        } else if (isLocalTask(childID)) {
          // toRemove.add(childID); TODO
          console.println(
                  workerID + " task.id = " + task.id + " is waiting for local child id = " + childID);
          pendingChildIDs.add(childID);
        }
      }
      for (long childID : toRemove) {
        childIDs.remove(childID);
      }

      if (!pendingChildIDs.isEmpty()) {
        this.pendingFrames.put(task.getID(), childIDs);
        this.pendingFramesLocal.put(task.getID(), pendingChildIDs);
        return;
      }
    }
    // frame is not waiting for local frames
    if (task.victims.size() > 0) { // this is not the first victim
      joinRemote(task, childIDs);
    } else { // task is ready for execution!
        synchronized (workerBag) {
          if (this.state == -2) {
            setState(0);
            final Bag bag = queueInitializer.get();
            bag.merge(task);
            async(
                () -> {
                  console.println(workerID + " run called from join at timestamp = " + (System.currentTimeMillis() - startTime));
                  run((B) bag);
                });
          } else if (this.state == -1) {
            workerBag.bag.merge(task);
          } else {
            workerBag.bag.merge(task);
          }
       }
    }
  }

  public void joinRemote(NFJTask task, ArrayList<Long> childIDs) {
    if (task.victims.size() > 0) { // frame was stolen from another !place!
      // (FORWARD) UNIFY PROTOCOL
      final int w = (int) task.victims.remove(task.victims.size() - 1);
      final NFJTask _task = task;
      final int _sourceID = workerID;
      computer.asyncAt(
          w,
          (GLBWorker worker) -> {
            final NFJTask t = (NFJTask) worker.givenFrames.remove(_task.getID());
            t.victims = _task.victims;
            if (t.victims.contains(w)) {
              worker.givenFrames.put(t.getID(), t);
            }
            t.mergeResults(_task);
            if (_task.getID() == 0) worker.console.println(worker.workerID + " root task moved (joinRemote) from worker " + _sourceID + "->" + worker.workerID);
            worker.join(t, childIDs);
          });
    } else {
      console.println(workerID + " NOT FROM ANOTHER PLACE?!?!?!? victims=" + task.victims);
    }
  }

  public void parentReturn(NFJTask task, Object result) {
    long childID = task.getID();
//    System.out.println(workerID + " Result return from task.id = " + task.id);
    NFJTask parentFrame = task.getParent();
    if (parentFrame == null) {
      System.out.println(workerID + " task=" + task.id + " PARENT RETURN WITH PARENT FRAME = NULL");
      System.out.println("Program took " + (System.currentTimeMillis() - startTime) + " ms");
      return; // program may end at this point
    }
    long parentID = parentFrame.getID();
    // store result locally (will be merged into parent frame on join)

    this.storedResults.put(childID, result);
    synchronized (this.pendingFrames) {
      if (this.givenFrames.containsKey(parentID)) {
        parentFrame = this.givenFrames.get(parentID);
      }
      if (parentFrame.isStolen()) {
        console.println(workerID + " parentReturn ending bc parent stolen task.id = " + task.id);
        return;
      }
      if (this.pendingFramesLocal.containsKey(parentID)) {
        ArrayList<Long> childIDs = this.pendingFrames.remove(parentID);
        this.pendingFramesLocal.remove(parentID);
        join(parentFrame, childIDs); // try to complete join again
      }
    }
//    System.out.println(workerID + " Parent return end task.id = " + task.id);
  }

  void answerLifelines() {
    console.println(
        "CALLED ANSWER LIFELINES! lifelineThieves.size="
            + lifelineThieves.size()
            + " "
            + "workerBag.bag.#tasks="
            + workerBag.bag.getCurrentTaskCount());
    while (!lifelineThieves.isEmpty() && workerBag.bag.getCurrentTaskCount() > 1 && workerBag.bag.isSplittable()) {
      int thief = lifelineThieves.poll();
      console.println("FOUND LIFELINE THIEF " + thief);
      console.println(
          "Stealnumtasks = " + GLBMultiWorkerConfiguration.GLB_MULTIWORKER_STEAL_NUMTASKS.get());
      B loot =
          workerBag.bag.split(
              false, GLBMultiWorkerConfiguration.GLB_MULTIWORKER_STEAL_NUMTASKS.get());
      openTasks = workerBag.bag.getCurrentTaskCount();
      latestWasSplittable = workerBag.bag.isSplittable();
      console.println("split loot size = " + loot.getCurrentTaskCount());
      for (NFJTask task : loot.asTasks()) {
        task.setStolen(true);
        this.givenFrames.put(task.getID(), task);
      }
      final int _workerID = workerID;
      console.println(workerID + " this = " + this);
      computer.asyncAt(
          thief,
          (GLBWorker worker) -> {
            worker.deal(_workerID, false, loot);
          });
    }
  }

  void answerRandomSteals() {
    final int victim = workerID;
    while (!randomStealRequests.isEmpty() && workerBag.bag.getCurrentTaskCount() > 1 && workerBag.bag.isSplittable()) {
      B loot =
          workerBag.bag.split(
              false, GLBMultiWorkerConfiguration.GLB_MULTIWORKER_STEAL_NUMTASKS.get());
      openTasks = workerBag.bag.getCurrentTaskCount();
      latestWasSplittable = workerBag.bag.isSplittable();
      int thief = randomStealRequests.poll();
      console.println(workerID + " sending random loot to thief = " + thief + "timestamp = " + (System.currentTimeMillis() - startTime));
      for (NFJTask task : loot.asTasks()) {
        task.setStolen(true);
        this.givenFrames.put(task.getID(), task);
      }
      computer.uncountedAsyncAt(
          thief,
          (GLBWorker worker) -> {
            worker.deal(victim, true, loot);
          });
    }
  }

  void rejectAll() {
    console.println(workerID + " Rejecting all");
    while (!randomStealRequests.isEmpty()) {
      int thief = randomStealRequests.poll();
      computer.uncountedAsyncAt(
          thief,
          (GLBWorker worker) -> {
            worker.console.println(worker.workerID + " received reject message (rejectAll)");
            worker.waitLatch.countDown();
          });
    }
  }

  /**
   * Main procedure of a worker thread in a place
   *
   * <p>A worker is a thread that processes the computation on a place. It has a {@link Bag}
   * instance to process and an identifier.
   *
   * @param workerBag computation to process along with an identifier for this worker process
   */
  void workerProcess(WorkerBag workerBag) {
    console.println("workerProcess on worker " + workerID);
    logger.workerStarted();
    openTasks = workerBag.bag.getCurrentTaskCount();
    latestWasSplittable = workerBag.bag.isSplittable();
    NFJTask currentTask = null;

    for (; ; ) { // Infinite loop, not a mistake
      boolean cont = true;
      do {
        /*
         * Checking if waiting lifelines can be answered
         */
        synchronized (workerBag) {
          /*
           * 6. Process its bag
           */
          if (currentTask != null) workerBag.bag.merge(currentTask);
          currentTask =
              workerBag.bag.process(
                  GLBMultiWorkerConfiguration.GLB_MULTIWORKER_N.get(), result, this);
          cont = !workerBag.bag.isEmpty() || currentTask != null;
          openTasks = workerBag.bag.getCurrentTaskCount();
          latestWasSplittable = workerBag.bag.isSplittable();
          if (cont && !randomStealRequests.isEmpty()) {
            answerRandomSteals();
          }
          if (cont && !lifelineThieves.isEmpty()) {
            answerLifelines();
          }
       }

      } while (cont); // 7. Repeat previous steps until the bag becomes empty.

      logger.workerStealing(); // The worker is now stealing

      setState(-1);
      rejectAll();
      console.println(workerID + " starting stealing at timestamp = " + (System.currentTimeMillis() - startTime));
      boolean success = doRandomSteals();
      synchronized (workerBag) {
        if (!success && workerBag.bag.isEmpty()) {
          doLifelineSteals();
          stopWorker(workerBag, -2);
          return;
        }
      }

      // Stealing from the queues in the place was successful. The worker goes
      // back to processing its fraction of the work.
    }
  }

  private void stopWorker(final WorkerBag workerBag, final int newState) {
    synchronized (workerBag) {
      if (state != -3) {
        setState(newState); // No more workers, we are now in stealing mode
      }
      rejectAll();
      logger.workerStopped();
      console.println(workerID + " Stopping worker");
    }
  }

  /**
   * Utility class used to contain a bag and the id of a worker in a single instance.
   *
   * @author Patrick Finnerty
   */
  class WorkerBag {

    /** Bag held by worker */
    public B bag;

    /** Integer identifier of the worker */
    public int workerId;

    /**
     * Constructor
     *
     * <p>Initializes a {@link WorkerBag} instance which holds the given id and {@link Bag}. This is
     * used to identify the workers
     *
     * @param id identifier of the worker that holds the bag
     * @param b Bag instance associated to the given identifier
     */
    public WorkerBag(int id, SerializableSupplier<B> b) {
      workerId = id;
      bag = b.get();
    }

    public void initStaticTasks() {
      bag.initStaticTasks(workerId);
    }
  }
}
