/*
 *  This file is part of the Handy Tools for Distributed Computing project
 *  HanDist (https://github.com/handist)
 *
 *  This file is licensed to You under the Eclipse Public License (EPL);
 *  You may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *      http://www.opensource.org/licenses/eclipse-1.0.php
 *
 *  (C) copyright CS29 Fine 2018-2019.
 */
package nfjglbsw.multiworker;

import static apgas.Constructs.*;

import apgas.Constructs;
import apgas.util.PlaceLocalObject;
import java.io.*;

/*
  Handles communication between workers by mapping worker numbers to GLBWorker objects.
*/
public class GLBcomputer<R extends Fold<R> & Serializable, B extends Bag<B, R> & Serializable>
    extends PlaceLocalObject {

  public transient GLBWorker workers[];

  public int getNumWorkers() {
    return workers.length * places().size();
  }

  public GLBWorker getWorker(int globalWorkerID) {
    int localWorkerID = globalWorkerID % workers.length;
    return this.workers[localWorkerID];
  }

  public boolean isLocalWorker(int globalWorkerID) {
    return here().id == (globalWorkerID / workers.length);
  }

  public void asyncAt(int globalWorkerID, SerializableConsumer<GLBWorker> j) {
    int placeID = globalWorkerID / workers.length;
    try {
      Constructs.asyncAt(
          place(placeID),
          () -> {
            j.accept(getWorker(globalWorkerID));
          });
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public void uncountedAsyncAt(int globalWorkerID, SerializableConsumer<GLBWorker> j) {
    int placeID = globalWorkerID / workers.length;
    try {
      Constructs.uncountedAsyncAt(
              place(placeID),
              () -> {
                j.accept(getWorker(globalWorkerID));
              });
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public GLBcomputer() {
    // create worker instances

    int numWorkers = GLBMultiWorkerConfiguration.GLB_MULTIWORKER_WORKERPERPLACE.get();
    workers = new GLBWorker[numWorkers];
    final int placeID = here().id;
    final int totalWorkers = numWorkers * places().size();
    finish(
        () -> {
          for (int i = 0; i < numWorkers; ++i) {
            final int workerID = numWorkers * placeID + i;
            final GLBWorker<R, B> worker = new GLBWorker<R, B>(workerID, this, totalWorkers);
            workers[i] = worker;
          }
        });
  }

  /**
   * Computes the given bag and returns the aggregated result of this computation.
   *
   * @param bag the computation to be performed
   * @param initResultSupplier function that provides new empty result instances
   * @param emptyBagSupplier function that provides new empty computation bag instances
   * @return aggregated result of the computation
   */
  public R computeDynamic(
      B bag, SerializableSupplier<R> initResultSupplier, SerializableSupplier<B> emptyBagSupplier) {
    resetAll(initResultSupplier, emptyBagSupplier);
    return (R)
        workers[0].computeDynamic(bag, initResultSupplier, emptyBagSupplier, emptyBagSupplier);
  }

  public void resetAll(SerializableSupplier initResult, SerializableSupplier emptyBag) {
    finish(
        () -> {
          for (int p = 0; p < places().size(); ++p) {
            for (int w = 0; w < workers.length; ++w) {
              final int globalWorkerID = p * workers.length + w;
              asyncAt(
                  globalWorkerID,
                  (GLBWorker worker) -> worker.reset(initResult, emptyBag, emptyBag));
            }
          }
        });
  }

  // returns global worker id
  public int getBestLocalWorker() {
    int bestID = -1;
    long bestN = 0;
    for (GLBWorker worker : workers) {
      long n = worker.openTasks;
      if (n > bestN && worker.latestWasSplittable) {
        bestID = worker.workerID;
        bestN = n;
      }
    }
    return bestID;
  }

  public void processIncomingSteal(int originalVictim, int thief, boolean lifeline) {
    int bestWorker = originalVictim;
    if (workers[0].localoptMode == 2 && !isLocalWorker(thief)) {
      int b = getBestLocalWorker();
      if (b != -1) {
        bestWorker = b;
      }
    }
    getWorker(bestWorker).processIncomingSteal(thief, lifeline);
  }
}
