package nfjglbswft.examples.syntheticBenchmark;

import nfjglbswft.multiworker.GLBMultiWorkerConfiguration;
import nfjglbswft.multiworker.GLBWorker;

import java.io.Serializable;
import java.lang.management.ManagementFactory;
import java.lang.management.ThreadMXBean;
import java.util.Random;

import static apgas.Constructs.here;
import static apgas.Constructs.places;

public class Synthetic implements Serializable {
  static long expectedResult;

  protected final SyntheticTaskDeque tasks;
  public long durationVariance;
  protected long count = 0;
  protected long result = 0;
  public long maxChildren = 0;
  public float variance = 0;

  protected transient Random randGen = new Random();
  public long diff = 0;
  transient ThreadMXBean bean = ManagementFactory.getThreadMXBean();

  /*
   *  dynamic initialization
   *  On initialization one task is generated on worker 0.
   *  Processing tasks can generate new tasks.
   */
  public Synthetic(final long durationVariance, final long maxChildren) {
    this(durationVariance, maxChildren, 64);
  }

  public Synthetic(final long durationVariance, final long maxChildren, final int initialCapacity) {
    tasks = new SyntheticTaskDeque(initialCapacity);
    this.durationVariance = durationVariance;
    this.variance = durationVariance / 100.0f;
    this.maxChildren = maxChildren;
  }

  long calculateTreeSize(long d, long c) {
    long taskCount = 1;
    for (long i = 1; i < d; ++i) {
      taskCount += Math.pow(c, i);
    }
    return taskCount;
  }

  long findTreeSize(long tasksPerWorker) {
    final int totalWorkers =
        GLBMultiWorkerConfiguration.GLB_MULTIWORKER_WORKERPERPLACE.get() * places().size();
    final long targetTotalTasks = tasksPerWorker * totalWorkers;
    long lastDiff = Long.MAX_VALUE;
    long lastDepth = 1;
    long lastChildren = 1;

    for (int d = 2; d < 24; ++d) {
      for (int c = 2; c < 8; ++c) {
        long currentTasks = calculateTreeSize(d, c);
        long currentDiff = Math.abs(currentTasks - targetTotalTasks);
        if (currentDiff < lastDiff) {
          lastDiff = currentDiff;
          lastDepth = d;
          lastChildren = c;
        }
      }
    }

    this.maxChildren = lastChildren;
    return lastDepth;
  }

  public long initDynamic(
      final long tasksPerWorker, final long totalDuration, final long taskBallast) {
    final int workerPerPlace = GLBMultiWorkerConfiguration.GLB_MULTIWORKER_WORKERPERPLACE.get();

    final int totalWorkers = workerPerPlace * places().size();
    final long localTasksPerWorker = tasksPerWorker;

    final long depth = findTreeSize(localTasksPerWorker);
    final long taskCount = calculateTreeSize(depth, maxChildren);
    final long taskDuration = (totalWorkers * totalDuration * 1000L * 1000L) / taskCount;
    tasks.addLast(new SyntheticTask(taskBallast, 0, depth - 1, taskDuration, (int) maxChildren, null, 0));
    expectedResult = taskCount;
    return taskCount;
  }

  public SyntheticTask calculate(SyntheticTask task, GLBWorker worker) {
    long before = bean.getCurrentThreadCpuTime();
    long after = bean.getCurrentThreadCpuTime();
    long beanDuration = after - before;

    randGen.setSeed(task.seed);
    if (task.depth > 0 && task.continuation < maxChildren) {
      SyntheticTask parentTask = task;
      tasks.addFirst(parentTask);
      task =
              new SyntheticTask(
                      task.ballast.length, task.seed + randGen.nextInt(), parentTask.depth - 1, parentTask.duration, (int) maxChildren, parentTask, worker.generateNewTaskID());
      parentTask.seed = parentTask.seed + randGen.nextInt();
      parentTask.childIDMap.put(task.getID(), parentTask.continuation);
      ++parentTask.continuation;
      long current = bean.getCurrentThreadCpuTime();
      task.processedTime += current - before;
      return task;
    } else if (task.continuation == maxChildren) {
      ++task.continuation;
      long childIDs[] = new long[(int) maxChildren];
      int idx = 0;
      for (Long childID : task.childIDMap.keySet()) {
        childIDs[idx++] = childID;
      }
      worker.join(task, childIDs);
      long current = bean.getCurrentThreadCpuTime();
      task.processedTime += current - before;
      return null;
    } else {
      long taskDuration;
      if (maxChildren > 0) {
        long x1 = (long) ((1 - variance) * task.duration);
        float randomFloat = randGen.nextFloat();
        long x2 = (long) (variance * task.duration * 2f * randomFloat);
        taskDuration = x1 + x2;
      } else {
        taskDuration = task.duration;
      }
      taskDuration -= diff;
      taskDuration -= beanDuration;

      long result = 1;
      for (int k = 0; k < task.childResults.length; ++k) {
        result += task.childResults[k];
      }

      long current2;
      long current = bean.getCurrentThreadCpuTime();
      task.processedTime += current - before;
      while (task.processedTime < taskDuration) {
        fibonacci(5);
        current2 = bean.getCurrentThreadCpuTime();
        task.processedTime += current2 - current;
        current = current2;
      }
      diff = task.processedTime - taskDuration;

      if (task.getParent() == null) {
        System.out.println(worker.workerID + " RESULT IS " + result);
        worker.setFinalResult(result);
        if (result != expectedResult) {
          System.out.println("RESULT NOT CORRECT! expected=" + expectedResult);
        }
      }
      worker.parentReturn(task, result);
      return null;
    }
  }

  private static int fibonacci(int a) {
    if (a == 1 || a == 2) return 1;
    else return fibonacci(a - 1) + fibonacci(a - 2);
  }
}
