package nfjglbswft.examples.uts;

import nfjglbswft.multiworker.Bag;
import nfjglbswft.multiworker.GLBWorker;
import nfjglbswft.multiworker.GLBcomputer;
import nfjglbswft.multiworker.NFJTask;

import java.io.Serializable;
import java.util.Iterator;

import static apgas.Constructs.getCurrentWorker;
import static apgas.Constructs.here;

public class UTS implements Bag<UTS, Long>, Serializable {

  protected static final double den = Math.log(4.0 / (1.0 + 4.0));

  public long cutOff;
  public UTSTaskQueue queue;

  public UTS(long cutOff) {
    this(cutOff, 64);
  }

  public UTS(long cutOff, int initialCapacity) {
    this.queue = new UTSTaskQueue(initialCapacity);
    this.cutOff = cutOff;
  }

  public void init(int seed, int depth, GLBcomputer<Long, UTS> computer) {
    UTSTask task = (new UTSTask(new SHA1Rand(seed, depth), 0, null, 0));
    int u = (int) Math.floor(Math.log(1 - task.getUTSParent().getRand() / 2147483648.0) / this.den);
    task.numChildren = u;
    task.continuation = u;
    task.childResults = new long[u];
    this.queue.addFirst(task);
    for (int i = 0; i < u; ++i) {
      UTSTask childTask = new UTSTask(task.getUTSParent(), i, task, i + 1);
      this.queue.addFirst(childTask);
      task.childIDMap.put(childTask.getID(), i);
    }
    computer.workers[0].setLastTaskID(u);
  }

  @Override
  public void initStaticTasks(int workerId) {
    // unused because dynamic
  }

  @Override
  public boolean isEmpty() {
    return this.queue.size() == 0;
  }

  @Override
  public boolean isSplittable() {
    for (UTSTask task : queue) {
      if (!task.reachedJoin()) {
        return this.queue.size() > 1;
      }
    }
    return false;
  }

  @Override
  public void merge(UTS other) {
    this.queue.pushArrayFirst(other.queue.toArray());
  }

  @Override
  public void merge(NFJTask task) {
    this.queue.addFirst((UTSTask) task);
  }

  @Override
  public NFJTask[] asTasks() {
    NFJTask[] tasks = new NFJTask[this.queue.size()];
    int i = 0;
    for (NFJTask task : this.queue) {
      tasks[i++] = task;
    }
    return tasks;
  }

  public long processSubTree(SHA1Rand parentRand, int utsId) { // DFS
    int depth = parentRand.getDepth();
    SHA1Rand sha1Rand = new SHA1Rand(parentRand, utsId, depth - 1);
    int numChildren = (int) Math.floor(Math.log(1 - sha1Rand.getRand() / 2147483648.0) / this.den);
    long result = numChildren;
    if (sha1Rand.getDepth() <= 1) {
      return numChildren;
    } else if (depth > 1) {
      for (int i = 0; i < numChildren; ++i) {
        result += processSubTree(sha1Rand, i);
      }
    }
    return result;
  }

  @Override
  public NFJTask process(int workAmount, GLBWorker<Long, UTS> computer) {
    int i = 0;
    if (this.queue.size() == 0) return null;
    UTSTask task = this.queue.removeFirst();
    while (i < workAmount) {
      ++i;
      if (task.continuation == 0) {
        task.continuation = 1;
        final UTSTask parentTask = task;
        int depth = task.getUTSParent().getDepth();
        if (depth > 1) {
          SHA1Rand sha1Rand = new SHA1Rand(task.getUTSParent(), task.getUTSID(), depth - 1);
          int u = (int) Math.floor(Math.log(1 - sha1Rand.getRand() / 2147483648.0) / this.den);
          task.numChildren = u;
          if (sha1Rand.getDepth() > 1 && u > 0) {
            if (depth <= cutOff) {
              long result = task.numChildren;
              for (int j = 0; j < task.numChildren; ++j) {
                result += processSubTree(sha1Rand, j);
              }
              computer.parentReturn(task, result);
              task = null;
            } else {
              this.queue.addFirst(task);
              task.childResults = new long[task.numChildren];
              task.sha1Rand = sha1Rand;
              task = new UTSTask(sha1Rand, 0, parentTask, computer.generateNewTaskID());
              parentTask.childIDMap.put(task.getID(), 0);
              continue;
            }
          } else {
            computer.parentReturn(task, (long) task.numChildren);
            task = null;
          }
        }
      } else if (task.childResults != null && task.continuation < task.numChildren) {
        this.queue.addFirst(task);
        final UTSTask parentTask = task;
        task =
            new UTSTask(
                parentTask.sha1Rand,
                parentTask.continuation,
                parentTask,
                computer.generateNewTaskID());
        parentTask.childIDMap.put(task.getID(), parentTask.continuation);
        ++parentTask.continuation;
        continue;
      } else if (task.childResults != null && task.continuation == task.numChildren) { // join
        ++task.continuation;
        long childIDs[] = new long[task.numChildren];
        int idx = 0;
        for (Long childID : task.childIDMap.keySet()) {
          childIDs[idx++] = childID;
        }
        computer.join(task, childIDs);
        task = null;
      } else { // return
        long result = task.numChildren;
        for (int k = 0; k < task.numChildren && task.childResults != null; ++k) {
          result += task.childResults[k];
        }

        if (task.getParent() == null) {
          ++result;
          System.out.println(here() + " RESULT IS " + result);
          computer.setFinalResult(result);
        }

        computer.parentReturn(task, result);
        task = null;
      }
      if (this.queue.size() > 0 && i < workAmount) {
        task = this.queue.removeFirst();
      } else {
        break;
      }
    }
    return task;
  }

  public UTS split(boolean takeAll, int numTasks) {
    UTS other = new UTS(cutOff, 1);

    if (!isSplittable() && takeAll) {
      other.queue.pushArrayFirst(this.queue.getFromLast(this.queue.size()));
      return other;
    }

    int joined = 0;
    for (Iterator<UTSTask> iterator = queue.descendingIterator(); iterator.hasNext(); ) {
      UTSTask task = iterator.next();
      if (!task.reachedJoin()) {
        UTSTask[] fromLast = this.queue.getFromLast(joined);
        other.queue.addLast(this.queue.removeLast());
        this.queue.pushArrayLast(fromLast);
        break;
      } else {
        joined++;
      }
    }
    return other;
  }

  @Override
  public long getCurrentTaskCount() {
    return this.queue.size();
  }

  @Override
  public UTS deepCopy() {
    UTS other = new UTS(cutOff);
    for (UTSTask task : this.queue) {
      other.queue.addLast(new UTSTask(task));
    }
    return other;
  }
}
