package nfjglbswft.multiworker;

import apgas.Configuration;
import apgas.DeadPlaceException;
import apgas.Place;
import apgas.impl.GlobalRuntimeImpl;
import apgas.util.ConsolePrinter;
import apgas.util.ExactlyOnceExecutor;
import apgas.util.IncrementalEntryValue;
import com.hazelcast.config.MapConfig;
import com.hazelcast.core.*;
import com.hazelcast.map.EntryBackupProcessor;
import com.hazelcast.map.EntryProcessor;
import com.hazelcast.transaction.TransactionOptions;
import com.hazelcast.transaction.TransactionalTaskContext;
import com.sun.jdi.event.ExceptionEvent;
import nfjglbswft.multiworker.lifeline.LifelineStrategy;

import java.io.Serializable;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicIntegerArray;
import java.util.concurrent.atomic.AtomicLong;

import static apgas.Constructs.*;

/**
 * Class {@link GLBWorker} implements a lifeline-based work-stealing scheme between hosts with an
 * internal management that allows multiple workers to perform computation concurrently on the same
 * Java process.
 *
 * <p>This {@link GLBWorker} handles distributed computation that fits the {@link Bag} interface
 * using multiple concurrent workers on the same hosts. It follows the design proposed by Yamashita
 * and Kamada in their article <a href=
 * "https://www.jstage.jst.go.jp/article/ipsjjip/24/2/24_416/_article">Introducing a Multithread and
 * Multistage Mechanism for the Global Load Balancing Library of X10</a>. The current design does
 * not implement the multi-stage mechanisms described in their article.
 *
 * <p>The requirements on the kind of computation and the features available to the programmer are
 * further detailed in the documentation of classes {@link Bag} and {@link Fold}. The customization
 * possibilities of the load balance algorithm are presented in class {@link
 * GLBMultiWorkerConfiguration}.
 *
 * <p>The work-stealing scheme is similar in spirit to the original lifeline-based Global Load
 * Balancer implemented in the X10 programming language. Inactive hosts passively wait for some work
 * to reach them through their lifelines to resume computation. The major difference with the
 * original scheme comes in the fact that this implementation accommodates for several concurrent
 * workers in a single Java process. Some load balance is performed internally to keep the workers
 * occupied as much as possible while still allowing remote hosts to steal work. The design relies
 * on two {@link Bag} instances that are kept aside to perform load balance. One is primarily in
 * charge of load balance operations between the workers running on the local host (member {@link
 * #intraPlaceQueue}) while the other in dedicated to steals from remote hosts (member {@link
 * #interPlaceQueue}).
 *
 * @author Patrick Finnerty
 */

/** <R> type of the result produced by the computation <B> type of the computation bag */
public class GLBWorker<R extends Serializable, B extends Bag<B, R> & Serializable> {

  /** Holds the lifelineStrategy */
  private final LifelineStrategy lifelineStrategy;
  private final int backupInterval;
  private final int crashCase;
  private final long[] crashTimes;
  private final int crashNumber;
  private final long crashSeed;
  private final int crashBegin;
  private final int crashEnd;
  private HashMap<Integer, Integer> workerKeyMap = null;
  private final int startPlacesSize;
  private final int startWorkersSize;

  /** Printing Helper */
  ConsolePrinter console = ConsolePrinter.getInstance();

  private final transient ExactlyOnceExecutor exactlyOnceExecutor = new ExactlyOnceExecutor();

  /**
   * Lifelines this place can establish.
   */
  int[] LIFELINE;
  /**
   * Places that can establish a lifeline on this place.
   */
  int[] REVERSE_LIFELINE;
  /**
   * {@link Logger} instance used to gather the {@link PlaceLogger}s of each place and hold runtime
   * information common to the whole computation.
   *
   * @see #getLog()
   */
  Logger computationLog;
  /**
   * Collection used to keep track of the lifelines this place has established on other places.
   *
   * <p>The key is the integer identifying the remote place on which this place may establish
   * lifelines, the value is {@code true} when the lifeline is established, {@code false} otherwise.
   * There is always a mapping in this member for every potential lifeline established by this
   * place.
   */
  ConcurrentHashMap<Integer, Boolean> lifelineEstablished;
  /**
   * Collection used to record the lifeline thieves that have requested some work from this place
   * but could not be given some work straight away as part of method steal.
   */
  ConcurrentLinkedQueue<Integer> lifelineThieves;
  /**
   * Logger instance used to log the runtime of the {@link GLBWorker} instance located at this place
   */
  PlaceLogger logger;
  /**
   * Indicates if the log aggregation has already being performed in method {@link #getLog()}.
   * Avoids a second aggregation of the logs if multiple calls to this method are made following a
   * computation.
   */
  boolean logsGiven;
  /** Random instance used to decide the victims of random steals. */
  Random random;
  /**
   * Instance in which the result of the computation performed at this place is going to be stored.
   * It is initialized with the given neutral element before the computation starts in method {@link
   * #reset(SerializableSupplier, SerializableSupplier, SerializableSupplier)}.
   */
  R result;
  /**
   * State of this worker.
   *
   * <ul>
   *   <li>0 running
   *   <li>-1 stealing
   *   <li>-2 inactive
   * </ul>
   */
  volatile int state;

  /** Global ID of this worker. */
  public final int workerID;

  private SerializableSupplier<R> resultInitializer;
  private SerializableSupplier<B> queueInitializer;
  private SerializableSupplier<B> workerInitializer;

  volatile CountDownLatch waitLatch;
  volatile int waitLatchTarget;


  ConcurrentLinkedQueue<Integer> randomStealRequests;

  volatile long lastTaskID;

  public long startTime;

  public long processedTime;

  public int[] randomWorkerArray;
  public int lastRandomWorkerIdx;

  public AtomicIntegerArray activeSteals;

  public volatile R finalResult = null;

  private final transient HazelcastInstance hz = Hazelcast.getHazelcastInstanceByName("apgas");
  protected final transient IMap<Integer, WorkerState> iMapBackup;
  protected final transient IMap<Integer, HashMap<Integer, Pair<Long, B>>> iMapOpenLoot;
  protected final transient IMap<Integer, HashMap<Integer, Tripel<Long, NFJTask, ArrayList<Long>>>> iMapOpenFrames;

  AtomicIntegerArray handlerDone;

  private final String iMapOpenLootName;
  private final String iMapBackupName;
  private final String iMapOpenFramesName;

  protected final String iMapBackupHandlerRemoveID;
  protected final String iMapOpenLootHandlerRemoveID;
  protected final String iMapOpenFramesHandlerRemoveID;

  int backupCount;
  private long workerStartTime;

  public void setFinalResult(R finalResult) {
    this.finalResult = finalResult;
  }

  public void setLastTaskID(long lastTaskID) {
    this.lastTaskID = lastTaskID;
  }

  public long generateNewTaskID() {
    return ++lastTaskID;
  }

  public boolean isLocalTask(long taskID) {
    int ownerWorkerID = (int) (taskID >> 32L);
    return (
      workerID == ownerWorkerID
      || (isDead(place(computer.getPlaceID(ownerWorkerID))) && workerID == getBackupWorkerOf(ownerWorkerID))
    );
  }

  public int getNextRandomVictim() {
    lastRandomWorkerIdx = (lastRandomWorkerIdx + 1) % randomWorkerArray.length;
    return randomWorkerArray[lastRandomWorkerIdx];
  }

  GLBcomputer<R, B> computer;


  static class WorkerState<R extends Serializable, B extends Bag<B, R> & Serializable> extends IncrementalEntryValue implements Serializable {
    boolean recovered;
    volatile WorkerBag<R, B> workerBag;
    ConcurrentHashMap<Long, NFJTask>
            givenFrames; // frames that were given away to a thief (TaskID -> Task)
    ConcurrentHashMap<Long, Object>
            storedResults; // results not yet inserted into the parent frame (TaskID -> TaskResult)
    ConcurrentHashMap<Long, ArrayList<Long>>
            pendingFrames; // frames that wait for task results (TaskID -> List(ChildTaskIDs))
    ConcurrentHashMap<Long, ArrayList<Long>>
            pendingFramesLocal; // frames that wait for local task results (TaskID -> List(ChildTaskIDs))

    ConcurrentHashMap<Long, NFJTask<R>> joinedTasks;

    final long[] receivedLids;
    long currentLid;
    final long[] receivedFrids;
    long currentFrid;

    ConsolePrinter console = ConsolePrinter.getInstance();

    public WorkerState(int totalWorkers) {
      storedResults = new ConcurrentHashMap<>();
      givenFrames = new ConcurrentHashMap<>();
      pendingFrames = new ConcurrentHashMap<>();
      pendingFramesLocal = new ConcurrentHashMap<>();
      joinedTasks = new ConcurrentHashMap<>();
      receivedLids = new long[totalWorkers];
      receivedFrids = new long[totalWorkers];
      currentLid = Long.MIN_VALUE + 1;
      Arrays.fill(receivedLids, Long.MIN_VALUE);
      currentFrid = Long.MIN_VALUE + 1;
      Arrays.fill(receivedFrids, Long.MIN_VALUE);
      recovered = false;
    }

    public NFJTask mergeTask(NFJTask<R> task) {
      if (givenFrames.containsKey(task.id)) {
        NFJTask<R> givenTask = givenFrames.get(task.id);
        if (givenTask.victims.size() < task.victims.size()) {
          givenTask.mergeResults(task);
          givenTask.victims = new ArrayList<>(task.victims);
        }
        if (givenTask.victims.size() > 0 && givenTask.victims.get(givenTask.victims.size() - 1) == workerBag.workerId) {
          givenTask.victims.remove(givenTask.victims.size() - 1);
        }
        givenTask.setStolen(false);
        if (givenTask.getParent() == null && task.getParent() != null) {
          givenTask.parentTask = mergeTask(task.getParent());
        }
        return givenTask;
      } else {
        if (task.victims.contains(workerBag.workerId)) {
          task.setStolen(false);
        }
        if (task.victims.size() > 0 && task.victims.get(task.victims.size() - 1) == workerBag.workerId) {
          task.victims.remove(task.victims.size() - 1);
        }
        return task;
      }
    }

    public void merge(WorkerState<R, B> other) {
      for (long taskID : other.givenFrames.keySet()) {
        NFJTask thisFrame = givenFrames.get(taskID);
        NFJTask otherFrame = other.givenFrames.get(taskID);
        if (thisFrame == null) {
          givenFrames.put(taskID, otherFrame);
        } else {
          if (otherFrame.parentTask != null && thisFrame.parentTask == null) { // otherFrame is frame from origin worker
            thisFrame.parentTask = mergeTask(otherFrame.parentTask);
          }
          if (thisFrame.victims.size() < otherFrame.victims.size()) { // otherFrame more up to date
            thisFrame.mergeResults(otherFrame);
            thisFrame.victims = new ArrayList<>(otherFrame.victims);
            if (thisFrame.victims.size() > 0 && (int)thisFrame.victims.get(thisFrame.victims.size() - 1) == workerBag.workerId) {
              thisFrame.victims.remove(thisFrame.victims.size() - 1);
              thisFrame.setStolen(false);
            }
          }
        }
      }
      NFJTask[] tasks = other.workerBag.bag.asTasks();
      for (int i = tasks.length - 1; i >= 0; --i) {
        workerBag.bag.merge(mergeTask(tasks[i]));
      }
      storedResults.putAll(other.storedResults);
      for (long taskID : other.pendingFrames.keySet()) {
        ArrayList<Long> thisList = pendingFrames.get(taskID);
        if (thisList == null) {
          pendingFrames.put(taskID, other.pendingFrames.get(taskID));
        } else {
          ArrayList<Long> otherList = other.pendingFrames.get(taskID);
          thisList.addAll(otherList);
        }
      }
      for (long taskID : other.pendingFramesLocal.keySet()) {
        ArrayList<Long> thisList = pendingFramesLocal.get(taskID);
        if (thisList == null) {
          pendingFramesLocal.put(taskID, other.pendingFramesLocal.get(taskID));
        } else {
          ArrayList<Long> otherList = other.pendingFramesLocal.get(taskID);
          thisList.addAll(otherList);
        }
      }
      for (NFJTask task : other.joinedTasks.values()) {
        joinedTasks.put(task.getID(), mergeTask(task));
        joinedTasks.get(task.getID()).setStolen(false);
      }
    }
  }

  final WorkerState<R, B> workerState;

  /** Constructor (package visibility) */
  GLBWorker(int workerID, GLBcomputer computer, int totalWorkers) {
    this.workerStartTime = -1;
    this.startPlacesSize = places().size();
    this.startWorkersSize = totalWorkers;
    this.startTime = System.currentTimeMillis();
    this.workerID = workerID;
    this.lastTaskID = 1 + (((long) workerID) << 32L);
    this.computer = computer;
    this.backupCount = Configuration.APGAS_BACKUPCOUNT.get();
    this.backupInterval = GLBMultiWorkerConfiguration.GLB_MULTIWORKER_FAULTTOLERANCE_CPINTERVAL.get();
    this.crashCase = GLBMultiWorkerConfiguration.GLB_MULTIWORKER_FAULTTOLERANCE_CRASHCASE.get();
    this.activeSteals = new AtomicIntegerArray(totalWorkers);
    for (int i = 0; i < totalWorkers; ++i) {
      this.activeSteals.set(i, 0);
    }

    handlerDone = new AtomicIntegerArray(totalWorkers);

    try {
      Set<Partition> partitions = hz.getPartitionService().getPartitions();
      HashMap<Integer, Integer> workerPartitionIdMap = new HashMap<>();
      for (int i = 0; i < totalWorkers; i++) {
        int placeID = computer.getPlaceID(i);
        Member member = GlobalRuntimeImpl.getRuntime().getMembers().get(placeID);
        int pID = -1;
        for (Partition p : partitions) {
          if (p.getOwner().getUuid().equals(member.getUuid())) {
            pID = p.getPartitionId();
          }
        }
        workerPartitionIdMap.put(i, pID);
      }

      this.workerKeyMap = new HashMap<>();
      int lastPartitionKey[] = new int[startPlacesSize];
      for (int i = 0; i < totalWorkers; i++) {
        int pID = workerPartitionIdMap.get(i);
        int key = 0;
        while (true) {
          try {
            Partition partition = hz.getPartitionService().getPartition(key);
            if (null == partition) {
              System.err.println(
                      "FTGLB constructor hz.getPartitionService().getPartition(" + key + ") is null!!!!");
            }

            if (null != partition && partition.getPartitionId() == pID) {
              if (lastPartitionKey[computer.getPlaceID(i)] < key) {
                lastPartitionKey[computer.getPlaceID(i)] = key;
                this.workerKeyMap.put(i, key);
                break;
              }
            }

          } catch (Throwable t) {
            t.printStackTrace();
          }
          key++;
        }
      }
    } catch(Throwable throwable) { throwable.printStackTrace(); }

    {
      iMapOpenLootName = "iMapOpenLoot";
      this.iMapOpenLoot = this.hz.getMap(iMapOpenLootName);
      this.iMapOpenLootHandlerRemoveID =
              this.iMapOpenLoot.addPartitionLostListener(
                      event -> {
                        System.out.println(
                                workerID
                                        + "(in partitionLostListener - iMapOpenLoot): "
                                        + event.toString()
                                        + " ...shutting down the cluster now, result will be wrong");
                        computer.killAll();
                      });

      MapConfig openLootMapConfig = new MapConfig(iMapOpenLootName);
      openLootMapConfig.setBackupCount(backupCount);
      this.hz.getConfig().addMapConfig(openLootMapConfig);

      if (workerID == 0) {
        for (int i = 0; i < this.startWorkersSize; i++) {
          HashMap<Integer, Pair<Long, B>> newHashMap = new HashMap<>();
          this.iMapOpenLoot.set(getBackupKey(i), newHashMap);
        }
      }
    }

    {
      iMapBackupName = "iMapBackup";
      this.iMapBackup = this.hz.getMap(iMapBackupName);
      this.iMapBackupHandlerRemoveID =
              this.iMapBackup.addPartitionLostListener(
                      event -> {
                        System.out.println(
                                workerID
                                        + "(in partitionLostListener - iMapBackup): "
                                        + event.toString()
                                        + " ...shutting down the cluster now, result will be wrong");
                        computer.killAll();
                      });

      MapConfig backupMapConfig = new MapConfig(iMapBackupName);
      backupMapConfig.setBackupCount(backupCount);
      this.hz.getConfig().addMapConfig(backupMapConfig);
    }

    {
      iMapOpenFramesName = "iMapOpenFrames";
      this.iMapOpenFrames = this.hz.getMap(iMapOpenFramesName);
      this.iMapOpenFramesHandlerRemoveID =
              this.iMapOpenFrames.addPartitionLostListener(
                      event -> {
                        System.out.println(
                                workerID
                                        + "(in partitionLostListener - iMapOpenFrames): "
                                        + event.toString()
                                        + " ...shutting down the cluster now, result will be wrong");
                        computer.killAll();
                      });

      MapConfig openFramesMapConfig = new MapConfig(iMapOpenFramesName);
      openFramesMapConfig.setBackupCount(backupCount);
      this.hz.getConfig().addMapConfig(openFramesMapConfig);

      if (workerID == 0) {
        for (int i = 0; i < this.startWorkersSize; i++) {
          HashMap<Integer, Tripel<Long, NFJTask, ArrayList<Long>>> newHashMap = new HashMap<>();
          this.iMapOpenFrames.set(getBackupKey(i), newHashMap);
        }
      }

      this.crashSeed = GLBMultiWorkerConfiguration.GLB_MULTIWORKER_FAULTTOLERANCE_CRASHSEED.get();
      this.crashNumber = GLBMultiWorkerConfiguration.GLB_MULTIWORKER_FAULTTOLERANCE_CRASHNUMBER.get();
      this.crashBegin = GLBMultiWorkerConfiguration.GLB_MULTIWORKER_FAULTTOLERANCE_CRASHBEGIN.get();
      this.crashEnd = GLBMultiWorkerConfiguration.GLB_MULTIWORKER_FAULTTOLERANCE_CRASHEND.get();

      Random randGen = new Random(this.crashSeed);
      this.crashTimes = new long[startPlacesSize];
      final long crashDuration = (crashEnd - crashBegin) * 1_000_000_000L;
      for (int i = 0; i < this.crashTimes.length; i++) {
      this.crashTimes[i] = Long.MAX_VALUE;
      }
      for (int i = 0; i < this.crashNumber; i++) {
        int crashingPlace = 0;
        do {
          crashingPlace = randGen.nextInt(startPlacesSize);
        } while (this.crashTimes[crashingPlace] != Long.MAX_VALUE);
        this.crashTimes[crashingPlace] =
                1_000_000_000L * crashBegin
                        + ((long) (i + 1) * ((long) crashDuration / (long) (this.crashNumber + 1)));
      }
    }

    LifelineStrategy ls = null;
    try {
      System.out.println(GLBMultiWorkerConfiguration.GLB_MULTIWORKER_LIFELINESTRATEGY.get());
      ls =
          (LifelineStrategy)
              Class.forName(GLBMultiWorkerConfiguration.GLB_MULTIWORKER_LIFELINESTRATEGY.get())
                  .newInstance();
    } catch (InstantiationException | IllegalAccessException | ClassNotFoundException e) {
      e.printStackTrace();
    }

    this.lifelineStrategy = ls;

    LIFELINE = lifelineStrategy.lifeline(workerID, totalWorkers);
    REVERSE_LIFELINE = lifelineStrategy.reverseLifeline(workerID, totalWorkers);

    random = new Random(workerID);

    lifelineEstablished = new ConcurrentHashMap<>(LIFELINE.length);
    lifelineThieves = new ConcurrentLinkedQueue<>();
    logger = new PlaceLogger(workerID);

    workerState = new WorkerState<>(totalWorkers);

    randomStealRequests = new ConcurrentLinkedQueue<>();

    // generate array containing all workers exactly once
    randomWorkerArray = new int[totalWorkers - 1];
    for (int i = 0; i < randomWorkerArray.length; ++i) {
      randomWorkerArray[i] = -1;
    }
    if (totalWorkers > 1) {
      for (int i = 0; i < totalWorkers; ++i) {
        if (i == workerID) continue;
        int idx = 0;
        do {
          idx = random.nextInt(randomWorkerArray.length);
        } while (randomWorkerArray[idx] != -1);
          randomWorkerArray[idx] = i;
      }
      lastRandomWorkerIdx = -1;
    }
  }

  public void placeFailureHandler(final Place deadPlace) {
    while (!this.hz.getPartitionService().isClusterSafe()) {
      try {
        Thread.sleep(100);
      } catch (InterruptedException e) {
        e.printStackTrace();
      }
    }
    if (crashCase == 13) {
      try {
        Thread.sleep(2000);
      } catch (Throwable throwable) {}
    }
    crashWorker(6, startWorkersSize - 1, "Crash of worker and its backup worker(s) in place failure handler", 0);
    synchronized (workerState) {
      try {
        int[] deadWorkers = computer.getWorkerIDs(deadPlace.id);
        for (int deadWorker : deadWorkers) {
          for (int i = 0; i < randomWorkerArray.length; ++i) {
            if (deadWorker == randomWorkerArray[i]) {
              randomWorkerArray[i] = 0; // FIXME dirty
            }
          }
          for (int i = 0; i < LIFELINE.length; ++i) {
            if (LIFELINE[i] == deadWorker) {
              LIFELINE[i] = 0;
            }
          }
          activeSteals.set(deadWorker, 0);
          randomStealRequests.removeIf((w) -> w == deadWorker);
          lifelineThieves.removeIf((w) -> w == deadWorker);

          if (waitLatch != null && waitLatchTarget == deadWorker) {
            waitLatch.countDown();
          }

          final int _workerID = workerID;
          final WorkerState<R, B> _workerState = workerState;
          TransactionOptions txOptions = new TransactionOptions().setTimeout(15, TimeUnit.SECONDS);

          // backup worker restores backup
            for (int currentDeadWorker = deadWorker;
               isDead(place(computer.getPlaceID(currentDeadWorker)));
               currentDeadWorker = (currentDeadWorker + startWorkersSize - 1) % startWorkersSize) {
            final int _currentDeadWorker = currentDeadWorker;
            int backupWorkerID = getBackupWorkerOf(currentDeadWorker);
            if (backupWorkerID == workerID) {
              // transaction set iMapBackup of dead worker to null and save own backup
              hz.executeTransaction(txOptions, (TransactionalTaskContext context) -> {
                try {
                  final TransactionalMap<Integer, WorkerState<R, B>> transBackupMap = context.getMap(iMapBackupName);
                  final TransactionalMap<Integer, HashMap<Integer, Pair<Long, B>>> transOpenLootMap = context.getMap(iMapOpenLootName);
                  final TransactionalMap<Integer, HashMap<Integer, Tripel<Long, NFJTask, ArrayList<Long>>>> transOpenFramesMap = context.getMap(iMapOpenFramesName);
                  HashMap<Integer, Pair<Long, B>> sentLootMap = transOpenLootMap.get(getBackupKey(_currentDeadWorker));
                  WorkerState<R, B> deadWorkerState = transBackupMap.get(getBackupKey(_currentDeadWorker));
                  if (deadWorkerState.recovered) { // already recovered
                    return null;
                  }

                  // merge state
                  workerState.merge(deadWorkerState);

                  for (int thief : sentLootMap.keySet()) {
                    if (isDead(place(computer.getPlaceID(thief)))) {
                      Pair<Long, B> loot = sentLootMap.get(thief);
                      if (loot != null) {
                        WorkerState<R, B> deadThiefState = transBackupMap.get(getBackupKey(thief));
                        if (loot.getFirst() > deadThiefState.receivedLids[_currentDeadWorker] && deadWorkerState.currentLid >= loot.getFirst()) { // loot not in victims backup, and also not in thiefs backup
                          for (NFJTask<R> task : loot.getSecond().asTasks()) {
                            task = workerState.mergeTask(task);
                            if (workerState.givenFrames.containsKey(task.getID())) {
                              workerState.givenFrames.put(task.getID(), task);
                            }
//                            task.victims.add(workerID);
                            task.setStolen(false);
                            workerState.workerBag.bag.merge(task);
                          }
                        }
                        sentLootMap.put(thief, null);
                      }
                    }
                  }
/*                  for (NFJTask task : workerState.givenFrames.values()) {
                    final ArrayList<Integer> victims = task.victims;
                    for (int i = 0; i < victims.size(); ++i) {
                      if (victims.get(i) == _currentDeadWorker) {
                        victims.set(i, backupWorkerID);
                      }
                    }
                  }*/
                  for (int p = 0; p < startWorkersSize; ++p) {
                    if (p == _currentDeadWorker) {
                      continue;
                    }
                    HashMap<Integer, Tripel<Long, NFJTask, ArrayList<Long>>> sentFramesMap = transOpenFramesMap.get(getBackupKey(p));
                    WorkerState<R, B> senderBackup = transBackupMap.get(getBackupKey(p));
                    Tripel<Long, NFJTask, ArrayList<Long>> tripel = sentFramesMap.get(_currentDeadWorker);
                    if (tripel != null) {
                      NFJTask<R> givenFrame = deadWorkerState.givenFrames.get(tripel.getSecond().getID());
                      if (givenFrame != null && senderBackup.currentFrid >= tripel.getFirst()) {
                        sentFramesMap.put(_currentDeadWorker, null);
                        joinRemoteActivity(tripel.getSecond(), backupWorkerID, p, tripel.getThird(), tripel.getFirst(), true);
                      }
                    }
                    transOpenFramesMap.set(getBackupKey(p), sentFramesMap);
                  }

                  deadWorkerState.recovered = true;
                  transBackupMap.set(getBackupKey(_currentDeadWorker), deadWorkerState);
                  transBackupMap.set(getBackupKey(workerID), workerState);
                  transOpenLootMap.set(getBackupKey(_currentDeadWorker), sentLootMap);
                } catch (Throwable throwable) {
                  throwable.printStackTrace();
                }

                return null;
              });

              HashMap<Integer, Pair<Long, B>> sentLoot = iMapOpenLoot.get(getBackupKey(_currentDeadWorker));
              WorkerState<R, B> backup = iMapBackup.get(getBackupKey(_currentDeadWorker));
              for (int thief : sentLoot.keySet()) {
                if (sentLoot.get(thief) == null) {
                  continue;
                }
                final B _loot = sentLoot.get(thief).getSecond();
                final long _lid = sentLoot.get(thief).getFirst();
                if (_lid > backup.currentLid) { // already merged above
                  continue;
                }
                try {
                  if (_workerID == thief) { // to prevent deadlock
                    deal(_currentDeadWorker, false, _loot, _lid);
                  } else {
                    computer.at(thief, (GLBWorker worker) -> {
                      try {
                        worker.deal(_currentDeadWorker, false, _loot, _lid);
                      } catch (Throwable e) {
                        e.printStackTrace();
                      }
                      return null;
                    });
                  }
                  sentLoot.put(thief, null);
                } catch (DeadPlaceException e) {
                  // merge loot
                  hz.executeTransaction(txOptions, (TransactionalTaskContext context) -> {
                    final TransactionalMap<Integer, WorkerState<R, B>> transBackupMap = context.getMap(iMapBackupName);
                    final TransactionalMap<Integer, HashMap<Integer, Pair<Long, B>>> transOpenLootMap = context.getMap(iMapOpenLootName);
                    WorkerState<R, B> deadBackup = transBackupMap.get(getBackupKey(_currentDeadWorker));
                    WorkerState<R, B> thiefBackup = transBackupMap.get(getBackupKey(thief));
                    if (_loot != null && _lid <= deadBackup.currentLid) { // loot is not in dead workers backup! MERGE
                      if (_loot != null && _lid > thiefBackup.receivedLids[_currentDeadWorker]) { // loot wasnt received yet!
                        for (NFJTask task : _loot.asTasks()) {
                          task.victims.add(_currentDeadWorker);
                          task.setStolen(false);
                          if (workerState.givenFrames.containsKey(task.getID())) {
                            NFJTask givenTask = workerState.givenFrames.get(task.getID());
                            workerState.givenFrames.put(task.getID(), task);
                            task.parentTask = givenTask.parentTask;
                          }
                          _workerState.workerBag.bag.merge(task);
                        }
                        //_workerState.receivedLids[deadWorker] = _lid;
                        sentLoot.put(thief, null);
                        transOpenLootMap.set(getBackupKey(_currentDeadWorker), sentLoot);
                        transBackupMap.set(getBackupKey(_workerID), _workerState);
                      }
                    }
                    return null;
                  });
                }
              }
              iMapOpenLoot.put(getBackupKey(_currentDeadWorker), sentLoot);

              // CASE: LOOT SENT FROM DEAD WORKER TO ANOTHER DEAD WORKER (who is unable to do it by itself...)
              // pick it up for the dead worker
              // included in above transaction!

              // CASE: FRAME RETURNS SENT TO DEAD WORKER MERGED ABOVE BY BACKUP WORKER
              // FIXME: SLOW!111
            }
          }


          // open loot cases:

          // CASE: LOOT SENT TO DEAD WORKER
          // pick it up again
          hz.executeTransaction(txOptions, (TransactionalTaskContext context) -> {
            final TransactionalMap<Integer, WorkerState<R, B>> transBackupMap = context.getMap(iMapBackupName);
            final TransactionalMap<Integer, HashMap<Integer, Pair<Long, B>>> transOpenLootMap = context.getMap(iMapOpenLootName);
            HashMap<Integer, Pair<Long, B>> sentLootMap = transOpenLootMap.get(getBackupKey(_workerID));
            Pair<Long, B> loot = sentLootMap.get(deadWorker);
            WorkerState<R, B> deadBackup = transBackupMap.get(getBackupKey(deadWorker));
            if (loot != null && loot.getFirst() > deadBackup.receivedLids[_workerID]) { // loot is not in dead workers backup
              for (NFJTask task : loot.getSecond().asTasks()) {
                NFJTask<R> givenFrame = workerState.givenFrames.remove(task.getID());
                if (givenFrame.victims.contains(workerID) || givenFrame.victims.stream().anyMatch((v) -> isDead(place(computer.getPlaceID(v))) && workerID == getBackupWorkerOf(v))) {
                  workerState.givenFrames.put(givenFrame.getID(), givenFrame);
                }
                givenFrame.setStolen(false);
                workerState.workerBag.bag.merge(givenFrame);
              }
              sentLootMap.put(deadWorker, null);
              transOpenLootMap.set(getBackupKey(_workerID), sentLootMap);
              transBackupMap.set(getBackupKey(_workerID), _workerState);
            }
            return null;
          });


          // open frames:

          // CASE: Every worker sees if there were frames sent to them and applies them if they werent received yet
          hz.executeTransaction(txOptions, (TransactionalTaskContext context) -> {
            final TransactionalMap<Integer, WorkerState<R, B>> transBackupMap = context.getMap(iMapBackupName);
            final TransactionalMap<Integer, HashMap<Integer, Tripel<Long, NFJTask, ArrayList<Long>>>> transOpenFramesMap = context.getMap(iMapOpenFramesName);
            HashMap<Integer, Tripel<Long, NFJTask, ArrayList<Long>>> sentFramesMap = transOpenFramesMap.get(getBackupKey(deadWorker));
            Tripel<Long, NFJTask, ArrayList<Long>> frame = sentFramesMap.get(_workerID);
            WorkerState<R, B> deadBackup = transBackupMap.get(getBackupKey(deadWorker));
            if (frame != null && frame.getFirst() <= deadBackup.currentFrid) { // dead backup does contain frame return (frame not in dead queue)
              if (workerState.givenFrames.containsKey(frame.getSecond().getID())) {
//              if (frame.getFirst() > _workerState.receivedFrids[deadWorker]) { // frame not applied yet
                final ArrayList<Integer> victims = frame.getSecond().victims;
/*                for (int i = 0; i < victims.size(); ++i) {
                  if (victims.get(i) == deadWorker) {
                    victims.set(i, getBackupWorkerOf(deadWorker));
                  }
                }*/
                transBackupMap.set(getBackupKey(_workerID), _workerState);
                sentFramesMap.put(_workerID, null);
                transOpenFramesMap.set(getBackupKey(deadWorker), sentFramesMap);
                joinRemoteActivity(frame.getSecond(), workerID, deadWorker, frame.getThird(), frame.getFirst());
              } else {
                sentFramesMap.put(_workerID, null);
                transOpenFramesMap.set(getBackupKey(deadWorker), sentFramesMap);
              }
            }
            return null;
          });

          // CASE: Every worker sees if it sent a return msg to the dead worker
          // idea: redirect to the backup worker (maybe queue them up so they get processed after the place failure handler)
          // already handled in an inefficient way by the backup worker (it reads all open frame entries and does it by itself)
          // CASE: frame returns sent from a dead worker to another dead worker
          // already handled by same code as above case

          handlerDone.set(deadWorker, 1);
          workerState.notifyAll();
        }
        for (NFJTask<R> task : workerState.joinedTasks.values()) {
          NFJTask<R> givenTask = workerState.givenFrames.get(task.getID());
          if (givenTask != null) {
            task = workerState.mergeTask(task);
          }
          join(task, workerState.pendingFrames.get(task.getID()), false);
        }
      } catch (Throwable t) {
        t.printStackTrace();
      }

      if (state < 0 && workerState.workerBag.bag.getCurrentTaskCount() > 0) {
        int oldState = state;
        setState(0);
        if (oldState == -1) {
          if (waitLatch != null) {
            waitLatch.countDown();
          }
        } else {
          try {
            async(
                    () -> {
                      run(workerState.workerBag.bag);
                    });
          } catch (Throwable throwable) {
            throwable.printStackTrace();
          }
        }
      }
    }
  }

  private int getBackupWorkerOf(int deadWorker) {
    int i = (deadWorker + 1) % startWorkersSize;
    while (true) {
      if (!isDead(new Place(computer.getPlaceID(i)))) {
        break;
      }
      i = (i + 1) % startWorkersSize;
    }
    return i;
  }

  private int getBackupKey(int workerID) {
    return this.workerKeyMap.get(workerID % startWorkersSize);
  }

  /**
   * Computes the given bag and returns the aggregated result of this computation.
   *
   * @param bag the computation to be performed
   * @param initResultSupplier function that provides new empty result instances
   * @param emptyBagSupplier function that provides new empty computation bag instances
   * @return aggregated result of the computation
   */
  public R computeDynamic(
          B bag, SerializableSupplier<R> initResultSupplier, SerializableSupplier<B> emptyBagSupplier) {
    return computeDynamic(bag, initResultSupplier, emptyBagSupplier, emptyBagSupplier);
  }

  /**
   * Computes the given bag and returns the result of the distributed computation.
   *
   * @param work initial work to be processed
   * @param resultInitializer initializer for the result instance
   * @param queueInitializer initializer for the queue used for load balancing purposes
   * @param workerInitializer initializer for the workers bag
   * @return instance of type R containing the result of the distributed computation
   */
  public R computeDynamic(
      B work,
      SerializableSupplier<R> resultInitializer,
      SerializableSupplier<B> queueInitializer,
      SerializableSupplier<B> workerInitializer) {
    // We reset every place
    final long initStart = System.nanoTime();
    this.resultInitializer = resultInitializer;
    this.queueInitializer = queueInitializer;
    this.workerInitializer = workerInitializer;

    // We launch the computation
    setState(0);
    final long start = System.nanoTime();
    run(work);
    while (finalResult == null) {
      try { Thread.sleep(10); } catch (Throwable throwable) {}
    }

    final long computationFinish = System.nanoTime();
    System.out.println("Process time: " + ((computationFinish - start) / 1E9) + " s");

    for (Place place : places()) {
      if (place == here()) {
        continue;
      }
      asyncAt(place, () -> {
        System.exit(0);
      });
    }
    System.exit(0);

    // Preparation for method getLog if it is called
    this.computationLog.setTimings(initStart, start, computationFinish);
    return finalResult;
  }

  /**
   * Method called on this worker when a victim of steal is answering and providing some loot.
   *
   * @param victim the id from worker sending the loot
   * @param random true if the steal is a random steal, otherwise false
   * @param loot the work that was stolen by this worker
   */
  void deal(final int victim, boolean random, B loot, final long victimLid) {
    if (random) {
      logger.stealsSuccess.incrementAndGet();
    } else {
      logger.lifelineStealsSuccess.incrementAndGet();
      lifelineEstablished.put(victim, false);
    }

    synchronized (workerState) {
      if (victimLid <= workerState.receivedLids[victim]) {
        String s = "";
        for (NFJTask ignoredTask : loot.asTasks()) {
          s += ignoredTask.getID() + ",";
        }
        return;
      }
      for (NFJTask task : loot.asTasks()) {
        task.victims.add(victim);
        task.setStolen(false);
        if (workerState.givenFrames.containsKey(task.getID())) {
          //      task.parentTask = null;
          NFJTask givenTask = workerState.givenFrames.get(task.getID());
          workerState.givenFrames.put(task.getID(), task);
          task.parentTask = givenTask.parentTask;
        }
      }
      workerState.receivedLids[victim] = victimLid;
      workerState.workerBag.bag.merge(loot);
      writeBackup(BACKUP_STEAL);

      final int thief = workerID;
      try {
        computer.uncountedAsyncAt(victim, (worker) -> {
          try {
            synchronized (worker.workerState) {
              HashMap<Integer, Pair<Long, B>> hashMap =
                      (HashMap<Integer, Pair<Long, B>>) worker.iMapOpenLoot.get(worker.getBackupKey(victim));
              Pair<Long, B> pair = hashMap.get(thief);
              if (pair != null && pair.getFirst() == victimLid) {
                hashMap.put(thief, null);
                worker.iMapOpenLoot.put(worker.getBackupKey(victim), hashMap);
              }
              worker.activeSteals.set(thief, 0);
            }
          } catch (Throwable throwable) {
            throwable.printStackTrace();
          }
        });
      } catch (Throwable throwable) {
        throwable.printStackTrace();
      }

      if (state == -2) {
        setState(0);
        try {
          async(
                  () -> {
                    run(workerState.workerBag.bag);
                  });
        } catch (Throwable throwable) {
          throwable.printStackTrace();
        }
      }
    }

    if (waitLatch != null && random == true) {
      waitLatch.countDown();
    }
  }

  public Logger getLog() {
    if (!logsGiven) {
      logsGiven = true;
    }
    return computationLog;
  }

  static final int BACKUP_REGULAR = 0;
  static final int BACKUP_STEAL = 1;
  static final int BACKUP_RETURN = 2;

  private static class RegularBackupProcessor implements EntryProcessor {
    public final WorkerState workerState;

    public RegularBackupProcessor(WorkerState workerState) {
      this.workerState = workerState;
    }

    @Override
    public Object process(Map.Entry entry) {
      boolean recovered = ((WorkerState) entry.getValue()).recovered;
      if (!recovered) {
        entry.setValue(workerState);
      }
      return null;
    }

    @Override
    public EntryBackupProcessor getBackupProcessor() {
      return this::process;
    }
  }

  void writeBackup(int type) {
    try {
      synchronized (workerState) {
        exactlyOnceExecutor.executeOnKey(
                iMapBackup,
                getBackupKey(workerID),
                new RegularBackupProcessor(workerState)
        );
      }
    } catch (Throwable throwable) {
      throwable.printStackTrace();
    }
  }

  /**
   * Resets the local GLBWorker instance to a ready to compute state in preparation for the next
   * computation. An instance of the result which the computation is going to produce is kept aside
   * for later use.
   *
   * <p>
   *
   * @param resultInitSupplier supplier of empty result instance
   * @param queueInitializer supplier of empty queues for load balancing purposes
   * @param workerInitializer supplier of empty bag for workers
   */
  void reset(
      SerializableSupplier<R> resultInitSupplier,
      SerializableSupplier<B> queueInitializer,
      SerializableSupplier<B> workerInitializer) {

    // Resetting the logger
    logger = new PlaceLogger(workerID);
    logsGiven = false;
    computationLog = new Logger();

    // Resetting the field used to keep the result
    result = resultInitSupplier.get();

    this.resultInitializer = resultInitSupplier;
    this.queueInitializer = queueInitializer;
    this.workerInitializer = workerInitializer;

    // Resetting flags
    setState(-2);

    workerState.workerBag = new WorkerBag<R, B>(workerID, workerInitializer);
    this.iMapBackup.set(getBackupKey(workerID), workerState);

    // We reset the established lifelines trackers
    final boolean lifelinesOn = (workerID != 0);
    for (final int i : LIFELINE) {
      lifelineEstablished.put(i, lifelinesOn);
    }

    // We establish lifelines on this place for initial work-stealing conditions
    for (final int i : REVERSE_LIFELINE) {
      // i==0 is normal case
      if (i == 0) {
        continue;
      }
      lifelineThieves.add(i);
    }
  }

  /**
   * Main procedure of a place.
   *
   * <p>Spawns the first worker thread (which will in turn recursively spawn other worker threads).
   * When all workers run out of work, attempts to steal work from remote hosts.
   *
   * @param b the initial bag to compute.
   */
  void run(B b) {

    // Prepare the first worker to process the work given as parameter
    if (b != null) { // dynamic computation
      synchronized (workerState) {
        final WorkerState<R, B> _workerState = workerState;
        if (_workerState.workerBag.bag != b) {
          _workerState.workerBag.bag.merge(b);
        }
      }
    }

    setState(0);
    workerProcess(workerState.workerBag); // Working
    // All the workers have stopped, this place does not have any work and
    // will now attempt to steal some from other places
  }

  // sends steal requests to a number of random victims
  // waits for the answers (either reject or loot)
  boolean doRandomSteals() {
    if (computer.getNumWorkers() == 1) return false;
    final int numTries = GLBMultiWorkerConfiguration.GLB_MULTIWORKER_W.get();
    final int thief = workerID;
    for (int i = 0; i < numTries; ++i) {
      int victim = getNextRandomVictim();

      if (waitLatch != null) {
        System.exit(22);
        waitLatch.countDown();
      }
      waitLatch = new CountDownLatch(1);
      waitLatchTarget = victim;
      try {
        computer.uncountedAsyncAt(
            victim,
            (GLBWorker worker) -> {
              try {
                worker.processIncomingSteal(thief, false);
              } catch (Throwable thr) { thr.printStackTrace(); }
            });
      } catch (Throwable t) {
        t.printStackTrace(System.out);
        waitLatch.countDown();
        waitLatch = null;
        continue;
      }

      long waitedFrom = System.currentTimeMillis();
      try {
        waitLatch.await();
      } catch (InterruptedException e) {
        e.printStackTrace();
      }
      long waitedTo = System.currentTimeMillis();
      waitLatch = null;

      // Checks if we have some work now
      boolean workAvailable;
      synchronized (workerState) {
       workAvailable = !workerState.workerBag.bag.isEmpty();
      }
      if (workAvailable) {
        setState(0);
        return true;
      }
    }
    return false;
  }

  // sends lifeline steal requests to this worker's lifeline buddies
  // does not wait
  void doLifelineSteals() {
    if (computer.getNumWorkers() == 1) return;
    final int thief = workerID;
    for (final int victim : LIFELINE) {
      try {
        computer.uncountedAsyncAt(
                victim,
                (GLBWorker worker) -> {
                  try {
                    worker.processIncomingSteal(thief, true);
                  } catch (Throwable thr) { thr.printStackTrace(); }
                });
      } catch (Throwable throwable) {
        throwable.printStackTrace();
      }
    }
  }

  long lastStateBegin;
  void setState(int state) {
    lastStateBegin = System.currentTimeMillis();
//    System.out.println(workerID + " setting state to " + state + " stackTrace = " + Arrays.toString(Thread.currentThread().getStackTrace()));
    this.state = state;
  }

  void processIncomingSteal(int thief, boolean lifeline) {
    final int victim = workerID;

    if (randomStealRequests.contains(thief)) {
      if (lifeline == false) {
        computer.uncountedAsyncAt(
                thief,
                (GLBWorker worker) -> {
                  try {
                    worker.waitLatch.countDown();
                  } catch (Throwable thr) { thr.printStackTrace(); }
                });
      }
      return;
    }
    if (lifelineThieves.contains(thief)) {
      if (lifeline == false) {
        if (state == 0) {
          lifelineThieves.remove(thief);
          randomStealRequests.add(thief);
        } else {
          computer.uncountedAsyncAt(
                  thief,
                  (GLBWorker worker) -> {
                    try {
                      worker.waitLatch.countDown();
                    } catch (Throwable thr) { thr.printStackTrace(); }
                  });
        }
      }
      return;
    }

    if (activeSteals.get(thief) == 1) { // makes sure loot arrives in order
      if (!lifeline) {
        computer.uncountedAsyncAt(
                thief,
                (GLBWorker worker) -> {
                  try {
                    worker.waitLatch.countDown();
                  } catch (Throwable thr) { thr.printStackTrace(); }
                });
      }
      return;
    }

    if (lifeline) {
      lifelineThieves.add(thief);
    } else {
      try {
        if (state != 0) { // reject
          computer.uncountedAsyncAt(
                  thief,
                  (GLBWorker worker) -> {
                    try {
                      worker.waitLatch.countDown();
                    } catch (Throwable thr) { thr.printStackTrace(); }
                  });
        } else {
          randomStealRequests.add(thief);
        }
      } catch (Exception e) {
        e.printStackTrace(System.out);
      }
    }
  }

  public void join(NFJTask task, long... childIDs) {
    ArrayList<Long> childList = new ArrayList();
    for (int i = 0; i < childIDs.length; ++i) {
      childList.add(childIDs[i]);
    }
    join(task, childList);
  }

  public void join(NFJTask task, ArrayList<Long> childIDs) {
    join(task, childIDs, false);
  }

  // this method returns immediately
  // the task will get scheduled on the first victim of this frame
  public void join(NFJTask task, ArrayList<Long> childIDs, boolean fromHandler) {
    // fill in locally calculated child results
    ArrayList<Long> pendingChildIDs = new ArrayList<>();
    ArrayList<Long> toRemove = new ArrayList<>();
    //synchronized (workerState.pendingFrames) {
    synchronized (workerState) {
      int ownerWorkerID = (int) (task.getID() >> 32L);
      if (!fromHandler && isLocalTask(task.getID()) && ownerWorkerID != workerID) {
        while (handlerDone.get(ownerWorkerID) != 1) {
          try {
            workerState.wait(100);
          } catch (InterruptedException e) {}
        }
      }
      for (long childID : childIDs) {
        if (workerState.storedResults.containsKey(childID)) {
          task.receiveChildResult(childID, workerState.storedResults.remove(childID));
          toRemove.add(childID);
        } else if (isLocalTask(childID)) {
          // toRemove.add(childID); TODO
          pendingChildIDs.add(childID);
          workerState.joinedTasks.put(task.id, task);
        }
      }
      for (long childID : toRemove) {
        childIDs.remove(childID);
      }

      if (!pendingChildIDs.isEmpty()) {
        workerState.pendingFrames.put(task.getID(), childIDs);
        workerState.pendingFramesLocal.put(task.getID(), pendingChildIDs);
        return;
      }
      for (int i = 0; i < task.victims.size(); ++i) {
        int vID = (int) task.victims.get(i);
        if (isDead(place(computer.getPlaceID(vID)))) {
          int newVID = getBackupWorkerOf(vID);
          task.victims.set(i, newVID);
        }
      }
      while (task.victims.size() > 0 && (Integer) task.victims.get(task.victims.size() - 1) == workerID) {
        task.victims.remove(task.victims.size() - 1);
      }
      // frame is not waiting for local frames
      workerState.joinedTasks.remove(task.id);
      if (task.victims.size() > 0) { // this is not the first victim
        joinRemote(task, childIDs);
      } else { // task is ready for execution!
        NFJTask<R> givenTask = workerState.givenFrames.get(task.getID());
        if (task.parentTask == null && givenTask != null) {
          task.parentTask = givenTask.parentTask;
        }
        synchronized (workerState) {
          if (this.state == -2) {
            setState(0);
            workerState.workerBag.bag.merge(task);
            try {
              async(
                      () -> {
                        run(workerState.workerBag.bag);
                      });
            } catch (Throwable throwable) {
              throwable.printStackTrace();
            }
          } else if (this.state == -1) {
            workerState.workerBag.bag.merge(task);
          } else {
            workerState.workerBag.bag.merge(task);
          }
        }
      }
    }
  }

  public void joinRemote(NFJTask task, ArrayList<Long> childIDs) {
    if (task.victims.size() > 0) { // frame was stolen from another !place!
      // (FORWARD) UNIFY PROTOCOL
      final int w = (int) task.victims.remove(task.victims.size() - 1);
      final NFJTask _task = task;
      final int _sourceID = workerID;
      workerState.currentFrid++;
      final long frid = workerState.currentFrid;

      try {
        {
          HashMap<Integer, Tripel<Long, NFJTask, ArrayList<Long>>> hashMap = iMapOpenFrames.get(getBackupKey(workerID));
          Tripel<Long, NFJTask, ArrayList<Long>> tripel = new Tripel<>(workerState.currentFrid, task, childIDs);
          hashMap.put(w, tripel);
          iMapOpenFrames.set(getBackupKey(workerID), hashMap);
          crashLastWorker(7,"In joinRemote after writing iMapOpenFrame but not iMapBackup", 0);
          writeBackup(BACKUP_RETURN);
          crashLastWorker(8,"In joinRemote after writing all backups but before sending the frame", 0);
        }
        computer.uncountedAsyncAt(
                w,
                (GLBWorker worker) -> {
                  if (worker.crashCase == 10 && _sourceID == worker.startWorkersSize - 1) {
                    try {
                      Thread.sleep(1000);
                      System.out.println("DELAYED MESSAGE SHOULD ARRIVE NOW! task.id = " + _task.getID() + " (joinRemote)");
                    } catch (Throwable t) {}
                  }
                  try {
                    worker.joinRemoteActivity(_task, w, _sourceID, childIDs, frid);
                  } catch (Throwable thr) { thr.printStackTrace(); }
                });
        crashLastWorker(10,"After sending joinRemote but delaying the receiving until after the restore", 0);
      } catch (Throwable t) {
        t.printStackTrace();
      }
    }
  }

  private void joinRemoteActivity(final NFJTask _task, final int w, final int _sourceID, ArrayList<Long> childIDs, final long frid) {
    joinRemoteActivity(_task, w, _sourceID, childIDs, frid, false); }

  private void joinRemoteActivity(final NFJTask _task, final int w, final int _sourceID, ArrayList<Long> childIDs, final long frid, boolean fromHandler) {
    synchronized (workerState) {
      int ownerWorkerID = (int) (_task.getID() >> 32L);
      if (!fromHandler && isLocalTask(_task.getID()) && ownerWorkerID != workerID) {
        while (handlerDone.get(ownerWorkerID) != 1) {
          try {
            workerState.wait(100);
          } catch (InterruptedException e) {}
        }
      }
      try {
        NFJTask<R> t = (NFJTask<R>) workerState.givenFrames.remove(_task.getID());
        if (t == null) {
          return;
        }
        t.victims = new ArrayList<>(_task.victims);
        t.stolen = false;
        if (t.victims.contains(w) || t.victims.stream().anyMatch((v) -> isDead(place(computer.getPlaceID(v))) && workerID == getBackupWorkerOf(v))) {
          workerState.givenFrames.put(t.getID(), t);
        }
        t.mergeResults(_task);
        if (t.getParent() == null && _task.getParent() != null) {
          t.parentTask = workerState.mergeTask(_task.getParent());
        }
        join(t, childIDs, fromHandler);
        workerState.receivedFrids[_sourceID] = frid;
        writeBackup(BACKUP_RETURN);
        try {
          if (!isDead(place(computer.getPlaceID(_sourceID)))) {
            computer.uncountedAsyncAt(_sourceID, (sourceWorker) -> {
              final GLBWorker<R, B> _sourceWorker = (GLBWorker) sourceWorker;
              synchronized (_sourceWorker.workerState) {
                HashMap<Integer, Tripel<Long, NFJTask, ArrayList<Long>>> hashMap = _sourceWorker.iMapOpenFrames.get(_sourceWorker.getBackupKey(_sourceID));
                Tripel<Long, NFJTask, ArrayList<Long>> tripel = hashMap.get(w);
                if (tripel != null && tripel.getFirst() == frid) {
                  hashMap.put(w, null);
                  _sourceWorker.iMapOpenFrames.set(_sourceWorker.getBackupKey(_sourceID), hashMap);
                }
              }
            });
          }
        } catch (Throwable throwable) {
          throwable.printStackTrace();
        }
      } catch (Throwable throwable) {
        System.out.print(workerID + " " + throwable);
        throwable.printStackTrace();
      }
    }
  }

  public void parentReturn(NFJTask task, Object result) {
    long childID = task.getID();
    NFJTask parentFrame = task.getParent();
    if (parentFrame == null) {
      System.out.println("last task id = " + task.getID());
      System.out.println("Program took " + (System.currentTimeMillis() - startTime) + " ms");
      return; // program may end at this point
    }
    long parentID = parentFrame.getID();
    // store result locally (will be merged into parent frame on join)

    synchronized (workerState) {
      int ownerWorkerID = (int) (task.getParent().getID() >> 32L);
      if (isLocalTask(task.getParent().getID()) && ownerWorkerID != workerID) {
        while (handlerDone.get(ownerWorkerID) != 1) {
          try {
            workerState.wait(100);
          } catch (InterruptedException e) {}
          Thread.dumpStack();
        }
      }
      workerState.storedResults.put(childID, result);
      if (workerState.givenFrames.containsKey(parentID)) {
        parentFrame = workerState.givenFrames.get(parentID);
      }
      if (workerState.joinedTasks.containsKey(parentID)) {
        parentFrame = workerState.joinedTasks.get(parentID);
      }
      if (parentFrame.isStolen()) {
        return;
      }
      if (workerState.pendingFramesLocal.containsKey(parentID)) {
        ArrayList<Long> childIDs = workerState.pendingFrames.remove(parentID);
        workerState.pendingFramesLocal.remove(parentID);
        join(parentFrame, childIDs, false); // try to complete join again
      }
    }
  }

  void answerLifelines() {
    while (!lifelineThieves.isEmpty() && workerState.workerBag.bag.getCurrentTaskCount() > 1 && workerState.workerBag.bag.isSplittable()) {
      int thief = lifelineThieves.poll();
      B loot =
          workerState.workerBag.bag.split(
              false, GLBMultiWorkerConfiguration.GLB_MULTIWORKER_STEAL_NUMTASKS.get());
      for (NFJTask task : loot.asTasks()) {
        NFJTask<R> givenTask = workerState.givenFrames.get(task.getID());
        if (task.parentTask == null && givenTask != null && givenTask.parentTask != null) {
          task = workerState.mergeTask(task);
        }
        task.setStolen(true);
        workerState.givenFrames.put(task.getID(), task);
      }
      give(thief, false, loot.deepCopy());
    }
  }

  void give(int thief, boolean random, B loot) {
    try {
      activeSteals.set(thief, 1);
      ++workerState.currentLid;
      final int _workerID = workerID;
      final long _currentLid = workerState.currentLid;
      Pair<Long, B> pair = new Pair<>(_currentLid,  loot);
      HashMap<Integer, Pair<Long, B>> hashMap = iMapOpenLoot.get(getBackupKey(workerID));
      hashMap.put(thief, pair);
      iMapOpenLoot.set(getBackupKey(workerID), hashMap);
      crashLastWorker(5,"Before sending tasks but with tasks in iMapOpenLoot and in iMapBackup (incomplete steal backup)", 0);
      writeBackup(BACKUP_STEAL);
      crashLastWorker(4,"Before sending tasks but after writing steal backups on victim", 0);
      computer.uncountedAsyncAt(
        thief,
        (GLBWorker worker) -> {
          if (worker.crashCase == 11 && _workerID == worker.startWorkersSize - 1) {
            try {
              Thread.sleep(1000);
              System.out.println("DELAYED MESSAGE SHOULD ARRIVE NOW!");
            } catch (Throwable t) {}
          }
          try {
            worker.deal(_workerID, random, loot, _currentLid);
          } catch (Throwable thr) { thr.printStackTrace();}
        });
    } catch (Throwable t) {
      t.printStackTrace();
    }
    crashLastWorker(11,"After sending loot but delaying the receiving until after the restore", 0);
    crashLastWorker(3,"After sending tasks and writing steal backups on victim", 0);
    crashLastWorker(13,"After sending tasks and writing steal backups on victim and delay recovery", 0);
  }

  void answerRandomSteals() {
    final int victim = workerID;
    while (!randomStealRequests.isEmpty() && workerState.workerBag.bag.getCurrentTaskCount() > 1 && workerState.workerBag.bag.isSplittable()) {
      B loot =
          workerState.workerBag.bag.split(
              false, GLBMultiWorkerConfiguration.GLB_MULTIWORKER_STEAL_NUMTASKS.get());
      int thief = randomStealRequests.poll();
      for (NFJTask task : loot.asTasks()) {
        task.setStolen(true);
        NFJTask<R> givenTask = workerState.givenFrames.get(task.getID());
        if (task.parentTask == null && givenTask != null && givenTask.parentTask != null) {
          task = workerState.mergeTask(task);
        }
        workerState.givenFrames.put(task.getID(), task);
      }
      give(thief, true, loot.deepCopy());
    }
  }

  void rejectAll() {
    while (!randomStealRequests.isEmpty()) {
      int thief = randomStealRequests.poll();
      try {
        computer.uncountedAsyncAt(
                thief,
                (GLBWorker worker) -> {
                  try {
                    worker.waitLatch.countDown();
                  } catch (Throwable thr) { thr.printStackTrace(); }
                });
      } catch (Throwable throwable) {
        throwable.printStackTrace();
      }
    }
  }

  void checkForErrors() {
    if (console.getStatus() == true) {
      HashSet<Long> taskIDs = new HashSet<>();
      for (NFJTask<R> task : workerState.workerBag.bag.asTasks()) {
        boolean unique = taskIDs.add(task.getID());
        if (!unique) {
          console.println(workerID + " HAS TASK DUPLICATION OF TASK.ID = " + task.getID() + " IN QUEUE: " + workerState.workerBag.bag);
        }
      }
      for (long taskID : workerState.givenFrames.keySet()) {
        if (workerState.givenFrames.get(taskID).getID() != taskID) {
          console.println(workerID + " HAS GIVENFRAMES CORRUPTION OF TASK.ID = " + taskID + " (WRONG TASK SAVED: " + workerState.givenFrames.get(taskID).getID() + ")");
        }
      }
    }
  }

  /**
   * Main procedure of a worker thread in a place
   *
   * <p>A worker is a thread that processes the computation on a place. It has a {@link Bag}
   * instance to process and an identifier.
   *
   * @param workerBag computation to process along with an identifier for this worker process
   */
  void workerProcess(WorkerBag workerBag) {
    if (workerStartTime == -1) {
      workerStartTime = System.nanoTime();
    }

    logger.workerStarted();
    NFJTask currentTask = null;
    crashLastWorker(2,"Exit after receiving tasks and worker start", 0);
    crashWorker(6,startWorkersSize - computer.workers.length - 1, "Crash of worker and its backup worker(s) at beginning of place failure handler", 0);

    long lastBackupTime = System.nanoTime() + (long) (backupInterval * 1E9 / startWorkersSize) * workerID;
    for (; ; ) { // Infinite loop, not a mistake
      boolean cont = true;
      do {
        /*
         * Checking if waiting lifelines can be answered
         */
        synchronized (workerState) {
          /*
           * 6. Process its bag
           */
          currentTask =
              workerBag.bag.process(
                  GLBMultiWorkerConfiguration.GLB_MULTIWORKER_N.get(), this);

          if (this.crashTimes[here().id] <= (System.nanoTime() - workerStartTime)) {
            System.out.println(this.logger.toString());
            System.out.println(
                    "Shutdown of place "
                            + here().id
                            + " at "
                            + System.nanoTime()
                            + ", elapsed since start: "
                            + (System.nanoTime() - workerStartTime)
                            + ", expected: "
                            + this.crashTimes[here().id]);
            System.exit(0);
          }

          cont = !workerBag.bag.isEmpty() || currentTask != null;
          if (currentTask != null) {
            workerBag.bag.merge(currentTask);
          }
          if (cont && !randomStealRequests.isEmpty()) {
            answerRandomSteals();
          }
          if (cont && !lifelineThieves.isEmpty()) {
            answerLifelines();
          }
        }
        crashLastWorker(9,"Random crash after some time", 12000);

        long currentTime = System.nanoTime();
        if ((currentTime - lastBackupTime) / 1E9 >= backupInterval) {
          writeBackup(BACKUP_REGULAR);
          crashLastWorker(1,"Exit after first regular backup", 0);
          lastBackupTime = currentTime;
        }
      } while (cont); // 7. Repeat previous steps until the bag becomes empty.

      logger.workerStealing(); // The worker is now stealing

      setState(-1);
      rejectAll();
      boolean success = doRandomSteals();
      synchronized (workerState) {
        if (!success && workerBag.bag.isEmpty()) {
          doLifelineSteals();
          stopWorker(workerBag, -2);
          return;
        }
      }

      // Stealing from the queues in the place was successful. The worker goes
      // back to processing its fraction of the work.
    }
  }

  private void crashLastWorker(int caseNumber, String text, int timeSinceStart) {
    crashWorker(caseNumber, startWorkersSize - 1, text, timeSinceStart);
  }

  // time in milliseconds
  private void crashWorker(int caseNumber, int targetWorker, String text, int timeSinceStart) {
    if (crashCase == caseNumber && targetWorker == workerID) {
      long currentTime = System.currentTimeMillis();
      if ((currentTime - startTime) > timeSinceStart) {
        System.out.println("CRASHING WORKER " + workerID + " (place: " + here().id + "): " + text + " CASE " + caseNumber);
        System.exit(caseNumber);
      }
    }
  }

  private void stopWorker(final WorkerBag workerBag, final int newState) {
    synchronized (workerState) {
      if (state != -3) {
        setState(newState); // No more workers, we are now in stealing mode
      }
      rejectAll();
      logger.workerStopped();
    }
  }

  /**
   * Utility class used to contain a bag and the id of a worker in a single instance.
   *
   * @author Patrick Finnerty
   */
  static class WorkerBag<R extends Serializable, B extends Bag<B, R> & Serializable> implements Serializable {

    /** Bag held by worker */
    public B bag;

    /** Integer identifier of the worker */
    public int workerId;

    /**
     * Constructor
     *
     * <p>Initializes a {@link WorkerBag} instance which holds the given id and {@link Bag}. This is
     * used to identify the workers
     *
     * @param id identifier of the worker that holds the bag
     * @param b Bag instance associated to the given identifier
     */
    public WorkerBag(int id, SerializableSupplier<B> b) {
      workerId = id;
      bag = b.get();
    }

    public void initStaticTasks() {
      bag.initStaticTasks(workerId);
    }
  }
}
