/*
 *  This file is part of the Handy Tools for Distributed Computing project
 *  HanDist (https://github.com/handist)
 *
 *  This file is licensed to You under the Eclipse Public License (EPL);
 *  You may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *      http://www.opensource.org/licenses/eclipse-1.0.php
 *
 *  (C) copyright CS29 Fine 2018-2019.
 */
package nfjglbswft.multiworker;

import apgas.Constructs;
import apgas.GlobalRuntime;
import apgas.Place;
import apgas.SerializableCallable;
import apgas.util.PlaceLocalObject;

import java.io.Serializable;

import static apgas.Constructs.*;

/*
  Handles communication between workers by mapping worker numbers to GLBWorker objects.
*/
public class GLBcomputer<R extends Serializable, B extends Bag<B, R> & Serializable>
    extends PlaceLocalObject {

  public transient GLBWorker workers[];

  public int getNumWorkers() {
    return workers.length * places().size();
  }

  public GLBWorker getWorker(int globalWorkerID) {
    int localWorkerID = getLocalWorkerID(globalWorkerID);
    return this.workers[localWorkerID];
  }

  public int getLocalWorkerID(int globalWorkerID) {
      return globalWorkerID % workers.length;
  }

  public int getPlaceID(int globalWorkerID) {
    return globalWorkerID / workers.length;
  }

  public boolean isLocalWorker(int globalWorkerID) {
    return here().id == (globalWorkerID / workers.length);
  }

  public int[] getWorkerIDs(int placeID) {
    int[] workerIDs = new int[workers.length];
    for (int i = 0; i < workers.length; ++i) {
      workerIDs[i] = workers.length * placeID + i;
    }
    return workerIDs;
  }

  public void asyncAt(int globalWorkerID, SerializableConsumer<GLBWorker> j) {
    int placeID = globalWorkerID / workers.length;
    try {
      Constructs.asyncAt(
          place(placeID),
          () -> {
            j.accept(getWorker(globalWorkerID));
          });
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public <T extends Serializable> T at(int globalWorkerID, SerializableFunction<GLBWorker, T> j) {
    int placeID = globalWorkerID / workers.length;
    try {
      return Constructs.at(
              place(placeID),
              () -> {
                return j.apply(getWorker(globalWorkerID));
              });
    } catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  public void uncountedAsyncAt(int globalWorkerID, SerializableConsumer<GLBWorker> j) {
    int placeID = globalWorkerID / workers.length;
    try {
      Constructs.uncountedAsyncAt(
              place(placeID),
              () -> {
                j.accept(getWorker(globalWorkerID));
              });
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public void placeFailureHandler(final Place deadPlace) {
    for (int i = 0; i < workers.length; ++i) {
      try {
        final int _i = i;
        async(() -> {
          workers[_i].placeFailureHandler(deadPlace);
        });
      } catch (Throwable throwable) {
        throwable.printStackTrace();
      }
    }
  }

  public GLBcomputer() {
    // create worker instances

    GlobalRuntime.getRuntime().setPlaceFailureHandler(this::placeFailureHandler);
    int numWorkers = GLBMultiWorkerConfiguration.GLB_MULTIWORKER_WORKERPERPLACE.get();
    workers = new GLBWorker[numWorkers];
    final int placeID = here().id;
    final int totalWorkers = numWorkers * places().size();
    finish(
        () -> {
          for (int i = 0; i < numWorkers; ++i) {
            final int workerID = numWorkers * placeID + i;
            final GLBWorker<R, B> worker = new GLBWorker<R, B>(workerID, this, totalWorkers);
            workers[i] = worker;
          }
        });
  }

  /**
   * Computes the given bag and returns the aggregated result of this computation.
   *
   * @param bag the computation to be performed
   * @param initResultSupplier function that provides new empty result instances
   * @param emptyBagSupplier function that provides new empty computation bag instances
   * @return aggregated result of the computation
   */
  public R computeDynamic(
          B bag, SerializableSupplier<R> initResultSupplier, SerializableSupplier<B> emptyBagSupplier) {
    resetAll(initResultSupplier, emptyBagSupplier);
    return (R)
        workers[0].computeDynamic(bag, initResultSupplier, emptyBagSupplier, emptyBagSupplier);
  }

  public void resetAll(SerializableSupplier initResult, SerializableSupplier emptyBag) {
    finish(
        () -> {
          for (int p = 0; p < places().size(); ++p) {
            for (int w = 0; w < workers.length; ++w) {
              final int globalWorkerID = p * workers.length + w;
              try {
                  asyncAt(
                          globalWorkerID,
                          (GLBWorker worker) -> worker.reset(initResult, emptyBag, emptyBag));
              } catch (Throwable t) {
                  t.printStackTrace();
              }
            }
          }
        });
  }

    public void killAll() {
      GlobalRuntime.getRuntime().setPlaceFailureHandler((deadPlace) -> System.exit(40));

      finish(
          () -> {
            for (Place place : places()) {
              if (place.id == here().id) {
                continue;
              }
              try {
                Constructs.asyncAt(
                    place,
                    () -> {
                      GlobalRuntime.getRuntime()
                          .setPlaceFailureHandler((deadPlace) -> System.exit(44));
                      System.exit(44);
                    });
              } catch (Throwable throwable) {
              }
            }
          });
      System.exit(44);
    }
}
